__all__ = ('StickerPack', )

from scarletio import export

from ...bases import DiscordEntity
from ...core import STICKER_PACKS
from ...http.urls import build_sticker_pack_banner_url, build_sticker_pack_banner_url_as
from ...precreate_helpers import process_precreate_parameters_and_raise_extra

from .fields import (
    parse_banner_id, parse_cover_sticker_id, parse_description, parse_id, parse_name, parse_sku_id, parse_stickers,
    put_banner_id, put_cover_sticker_id, put_description, put_id, put_name, put_sku_id,
    put_stickers, validate_banner_id, validate_cover_sticker_id, validate_description, validate_id, validate_name,
    validate_sku_id, validate_stickers
)


PRECREATE_FIELDS = {
    'banner_id': ('banner_id', validate_banner_id),
    'cover_sticker': ('cover_sticker_id', validate_cover_sticker_id),
    'cover_sticker_id': ('cover_sticker_id', validate_cover_sticker_id),
    'description': ('description', validate_description),
    'name': ('name', validate_name),
    'sku_id': ('sku_id', validate_sku_id),
    'stickers': ('stickers', validate_stickers),
}


@export
class StickerPack(DiscordEntity, immortal = True):
    """
    A sticker's pack.
    
    Attributes
    ----------
    banner_id : `int`
        The banner asset identifier of the sticker pack.
    cover_sticker_id : `int`
        The sticker's identifier, which the pack uses as it's banner. Defaults to `0` if not applicable.
    description : `str`
        The pack's description.
    id : `int`
        The sticker pack's identifier.
    name : `str`
        The name of the sticker pack.
    sku_id : `int`
        The Stock Keeping Unit identifier of the sticker pack.
    stickers : `None`, `frozenset` of ``Sticker``
        The stickers of the pack.
    """
    __slots__ = ('banner_id', 'cover_sticker_id', 'description', 'name', 'sku_id', 'stickers')
    
    
    def __new__(
        cls,
        *,
        banner_id = ...,
        cover_sticker_id = ...,
        description = ...,
        name = ...,
        sku_id = ...,
        stickers = ...,
    ):
        """
        Creates a new partial sticker pack.
        
        Parameters
        ----------
        banner_id : `int`, Optional (Keyword only)
            The banner asset identifier of the sticker pack.
        cover_sticker_id : `int`, ``Sticker``, Optional (Keyword only)
            The sticker's identifier, which the pack uses as it's banner.
        description : `None`, `str`, Optional (Keyword only)
            The pack's description.
        name : `str`, Optional (Keyword only)
            The name of the sticker pack.
        sku_id : `int`, Optional (Keyword only)
            The Stock Keeping Unit identifier of the sticker pack.
        stickers : `None`, `iterable` of ``Sticker``, Optional (Keyword only)
            The stickers of the pack.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # banner_id
        if banner_id is ...:
            banner_id = 0
        else:
            banner_id = validate_banner_id(banner_id)
        
        # cover_sticker_id
        if cover_sticker_id is ...:
            cover_sticker_id = 0
        else:
            cover_sticker_id = validate_cover_sticker_id(cover_sticker_id)
        
        # description
        if description is ...:
            description = None
        else:
            description = validate_description(description)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # sku_id
        if sku_id is ...:
            sku_id = 0
        else:
            sku_id = validate_sku_id(sku_id)
        
        # stickers
        if stickers is ...:
            stickers = None
        else:
            stickers = validate_stickers(stickers)
        
        # Construct
        
        self = object.__new__(cls)
        self.banner_id = banner_id
        self.cover_sticker_id = cover_sticker_id
        self.description = description
        self.id = 0
        self.name = name
        self.sku_id = sku_id
        self.stickers = stickers
        return self
    
    
    @classmethod
    def from_data(cls, data, *, force_update = False):
        """
        Creates a new ``StickerPack`` from the received data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            sticker-pack data.
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the sticker should be updated.
        
        Returns
        -------
        self : `instance<cls>`
        """
        sticker_pack_id = parse_id(data)
        
        try:
            self = STICKER_PACKS[sticker_pack_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = sticker_pack_id
            self._set_attributes(data)
            STICKER_PACKS[sticker_pack_id] = self
        
        else:
            if force_update:
                self._set_attributes(data)
        
        return self
    
    
    def _set_attributes(self, data):
        """
        Sets the sticker pack's attributes.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Sticker-pack data.
        """
        self.banner_id = parse_banner_id(data)
        self.cover_sticker_id = parse_cover_sticker_id(data)
        self.description = parse_description(data)
        self.name = parse_name(data)
        self.sku_id = parse_sku_id(data)
        self.stickers = parse_stickers(data)
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the sticker pack to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_banner_id(self.banner_id, data, defaults)
        put_cover_sticker_id(self.cover_sticker_id, data, defaults)
        put_description(self.description, data, defaults)
        put_id(self.id, data, defaults)
        put_name(self.name, data, defaults)
        put_sku_id(self.sku_id, data, defaults)
        put_stickers(self.stickers, data, defaults)
        return data
    
    
    @classmethod
    def precreate(cls, sticker_pack_id, **keyword_parameters):
        """
        Precreates the sticker pack with the given parameters. When the sticker pack is loaded, the precreated one
        will be picked up and its fields will be populated.
        
        > Since sticker-packs are not bound to any other objects, it cannot be determined whether they are up-to-date.
        > Meaning their attributes are not locked if the sticker is already loaded and can be changed with `.precreate`.
        
        Parameters
        ----------
        sticker_pack_id : `int`
            The sticker pack's id.
        **keyword_parameters : keyword parameters
            Additional predefined attributes for the sticker pack.
        
        Other Parameters
        ----------------
        banner_id : `int`, Optional (Keyword only)
            The banner asset identifier of the sticker pack.
        cover_sticker : `int`, ``Sticker``, Optional (Keyword only)
            Alternative for `cover_sticker_id`.
        cover_sticker_id : `int`, ``Sticker``, Optional (Keyword only)
            The sticker's identifier, which the pack uses as it's banner.
        description : `None`, `str`, Optional (Keyword only)
            The pack's description.
        name : `str`, Optional (Keyword only)
            The name of the sticker pack.
        sku_id : `int`, Optional (Keyword only)
            The Stock Keeping Unit identifier of the sticker pack.
        stickers : `None`, `iterable` of ``Sticker``, Optional (Keyword only)
            The stickers of the pack.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - A parameter's type is incorrect.
            - Extra parameters given.
        ValueError
            - A parameter's value is incorrect.
        """
        sticker_pack_id = validate_id(sticker_pack_id)

        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        try:
            self = STICKER_PACKS[sticker_pack_id]
        except KeyError:
            self = cls._create_empty(sticker_pack_id)
            STICKER_PACKS[sticker_pack_id] = self
        else:
            if not self.partial:
                return self
        
        if (processed is not None):
            for name, value in processed:
                setattr(self, name, value)
        
        return self
    
    
    @classmethod
    def _create_empty(cls, sticker_pack_id):
        """
        Creates a sticker pack with its default field values set.
        
        Parameters
        ----------
        sticker_pack_id : `int`
            The identifier of the sticker pack.
        
        Returns
        -------
        sticker_pack : `instance<cls>`
        """
        self = object.__new__(cls)
        self.banner_id = 0
        self.cover_sticker_id = 0
        self.description = None
        self.id = sticker_pack_id
        self.name = ''
        self.sku_id = 0
        self.stickers = None
        return self
    
    
    def __repr__(self):
        """Returns the sticker pack's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        emoji_id = self.id
        if emoji_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(self.id))
            repr_parts.append(',')
        
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two sticker packs are equal"""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two sticker packs are not equal"""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two sticker packs are equal. Type of `other` must match type of `self`.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other sticker ack to compare self to.
        
        Returns
        -------
        is_equal : `bool`
        """
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            return (self_id == other_id)
        
        # banner_id
        if self.banner_id != other.banner_id:
            return False
        
        # cover_sticker_id
        if self.cover_sticker_id != other.cover_sticker_id:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # sku_id
        if self.sku_id != other.sku_id:
            return False
        
        # stickers
        if self.stickers != other.stickers:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the sticker pack's hash."""
        emoji_id = self.id
        if emoji_id:
            return emoji_id
        
        return self._get_hash_partial()
    
    
    def _get_hash_partial(self):
        """
        Returns a partial sticker pack's hash value.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # banner_id
        hash_value ^= hash(self.banner_id)
        
        # cover_sticker_id
        hash_value ^= hash(self.cover_sticker_id)
        
        # description
        description = self.description
        if (description is not None):
            hash_value ^= hash(description)
        
        # name
        hash_value ^= hash(self.name)
        
        # sku_id
        hash_value ^= hash(self.sku_id)
        
        # stickers
        stickers = self.stickers
        if (stickers is not None):
            hash_value ^= len(stickers)
            
            for sticker in stickers:
                hash_value ^= hash(sticker)
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the sticker pack.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.banner_id = self.banner_id
        new.cover_sticker_id = self.cover_sticker_id
        new.description = self.description
        new.id = 0
        new.name = self.name
        new.sku_id = self.sku_id
        stickers = self.stickers
        if (stickers is not None):
            stickers = frozenset(iter(stickers))
        new.stickers = stickers
        return new
    
    
    def copy_with(
        self,
        *,
        banner_id = ...,
        cover_sticker_id = ...,
        description = ...,
        name = ...,
        sku_id = ...,
        stickers = ...,
    ):
        """
        Copies the sticker pack with the given fields.
        
        Parameters
        ----------
        banner_id : `int`, Optional (Keyword only)
            The banner asset identifier of the sticker pack.
        cover_sticker_id : `int`, ``Sticker``, Optional (Keyword only)
            The sticker's identifier, which the pack uses as it's banner.
        description : `None`, `str`, Optional (Keyword only)
            The pack's description.
        name : `str`, Optional (Keyword only)
            The name of the sticker pack.
        sku_id : `int`, Optional (Keyword only)
            The Stock Keeping Unit identifier of the sticker pack.
        stickers : `None`, `iterable` of ``Sticker``, Optional (Keyword only)
            The stickers of the pack.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # banner_id
        if banner_id is ...:
            banner_id = self.banner_id
        else:
            banner_id = validate_banner_id(banner_id)
        
        # cover_sticker_id
        if cover_sticker_id is ...:
            cover_sticker_id = self.cover_sticker_id
        else:
            cover_sticker_id = validate_cover_sticker_id(cover_sticker_id)
        
        # description
        if description is ...:
            description = self.description
        else:
            description = validate_description(description)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # sku_id
        if sku_id is ...:
            sku_id = self.sku_id
        else:
            sku_id = validate_sku_id(sku_id)
        
        # stickers
        if stickers is ...:
            stickers = self.stickers
            if (stickers is not None):
                stickers = frozenset(iter(stickers))
        else:
            stickers = validate_stickers(stickers)
        
        # Construct
        
        new = object.__new__(type(self))
        new.banner_id = banner_id
        new.cover_sticker_id = cover_sticker_id
        new.description = description
        new.id = 0
        new.name = name
        new.sku_id = sku_id
        new.stickers = stickers
        return new
    
    
    @property
    def partial(self):
        """
        Returns whether the sticker pack is partial.
        
        Returns
        -------
        partial : `bool`
        """
        return self.id == 0
    
    
    def iter_stickers(self):
        """
        Iterates over the stickers of the sticker pack.
        
        This method is an iterable generator.
        
        Yields
        ------
        sticker : ``Sticker``
        """
        stickers = self.stickers
        if (stickers is not None):
            yield from stickers
    
    
    def has_sticker(self, sticker):
        """
        Returns whether the sticker pack contains the given sticker.
        
        Parameters
        ----------
        sticker : ``Sticker``
            The sticker to check for.
        
        Returns
        -------
        has_sticker : `bool`
        """
        stickers = self.stickers
        if (stickers is None):
            return False
        
        return (sticker in stickers)
    
    
    @property
    def banner_url(self):
        """
        Returns the sticker pack's banner's url. If the sticker pack has no banner, then returns `None`.
        
        Returns
        -------
        url : `None | str`
        """
        return build_sticker_pack_banner_url(self.banner_id)
    
    
    def banner_url_as(self, ext = None, size = None):
        """
        Returns the sticker pack's banner's url. If the sticker pack has no banner, then returns `None`.
        
        Parameters
        ----------
        ext : `None | str` = `None`, Optional
            The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
        
        size : `None | int` = `None`, Optional
            The preferred minimal size of the image's url.
        
        Returns
        -------
        url : `None | str`
        """
        return build_sticker_pack_banner_url_as(self.banner_id, ext, size)



__all__ = ('ComponentMetadataAttachmentInput', )

from scarletio import copy_docs

from ..shared_helpers import create_auto_custom_id

from .base import ComponentMetadataBase
from .constants import MAX_VALUES_DEFAULT, MIN_VALUES_DEFAULT
from .fields import (
    parse_custom_id, parse_max_values, parse_min_values, parse_required, put_custom_id, put_max_values, put_min_values,
    put_required, validate_custom_id, validate_max_values__attachment_input, validate_min_values__attachment_input,
    validate_required
)


class ComponentMetadataAttachmentInput(ComponentMetadataBase):
    """
    Attachment input component metadata.
    
    Attributes
    ----------
    custom_id : `None | str`
        Custom identifier to detect which component was used by the user.
    
    max_values : `int
        The maximal amount of attachments to input.
    
    min_values : `int`
        The minimal amount of attachments to input.
    
    required : `bool`
        Whether the field is required to be fulfilled.
    """
    __slots__ = ('custom_id', 'max_values', 'min_values', 'required')
    
    
    def __new__(
        cls,
        *,
        custom_id = ...,
        max_values = ...,
        min_values = ...,
        required = ...,
    ):
        """
        Creates a new attachment input component metadata with the given parameters.
        
        Parameters
        ----------
        custom_id : `None | str`, Optional (Keyword only)
            Custom identifier to detect which component was used by the user.
        
        max_values : `int, Optional (Keyword only)
            The maximal amount of attachments to input.
        
        min_values : `int`, Optional (Keyword only)
            The minimal amount of attachments to input.
        
        required : `None | bool`, Optional (Keyword only)
            Whether the field is required to be fulfilled.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # custom_id
        if custom_id is ...:
            custom_id = None
        else:
            custom_id = validate_custom_id(custom_id)
        
        # max_values
        if max_values is ...:
            max_values = MAX_VALUES_DEFAULT
        else:
            max_values = validate_max_values__attachment_input(max_values)
        
        # min_values
        if min_values is ...:
            min_values = MIN_VALUES_DEFAULT
        else:
            min_values = validate_min_values__attachment_input(min_values)
        
        # required
        if required is ...:
            required = None
        else:
            if (required is not None):
                required = validate_required(required)
        
        # Auto detect required if not-given / None
        
        if (required is None):
            if min_values > 0:
                required = True
            else:
                required = False
        
        # Extra checks
        
        if custom_id is None:
            custom_id = create_auto_custom_id()
        
        # Construct
        self = object.__new__(cls)
        self.custom_id = custom_id
        self.max_values = max_values
        self.min_values = min_values
        self.required = required
        return self
    
    
    @classmethod
    @copy_docs(ComponentMetadataBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            custom_id = keyword_parameters.pop('custom_id', ...),
            max_values = keyword_parameters.pop('max_values', ...),
            min_values = keyword_parameters.pop('min_values', ...),
            required = keyword_parameters.pop('required', ...),
        )
    
    
    @copy_docs(ComponentMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', type(self).__name__]
        
        # System fields : custom_id
        
        # custom_id
        repr_parts.append(' custom_id = ')
        repr_parts.append(repr(self.custom_id))
        
        # Optional descriptive fields: min_values & max_values & required
        
        # min_values
        min_values = self.min_values
        if min_values != MIN_VALUES_DEFAULT:
            repr_parts.append(', min_values = ')
            repr_parts.append(repr(min_values))
        
        # max_values
        max_values = self.max_values
        if max_values != MAX_VALUES_DEFAULT:
            repr_parts.append(', max_values = ')
            repr_parts.append(repr(max_values))
        
        # required (relation with `min_values`)
        required = self.required
        if (min_values > 0) ^ required:
            repr_parts.append(', required = ')
            repr_parts.append(repr(required))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ComponentMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            hash_value ^= hash(custom_id)
        
        # max_values
        max_values = self.max_values
        if (max_values != 1):
            hash_value ^= (max_values << 18)
        
        # min_values
        min_values = self.min_values
        if (min_values != 1):
            min_values ^= (min_values << 22)
        
        # required
        if self.required:
            hash_value ^= (1 << 28)
        
        return hash_value
    
    
    @copy_docs(ComponentMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # custom_id
        if self.custom_id != other.custom_id:
            return False
        
        # max_values
        if self.max_values != other.max_values:
            return False
        
        # min_values
        if self.min_values != other.min_values:
            return False
        
        # required
        if self.required != other.required:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ComponentMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.custom_id = parse_custom_id(data)
        self.max_values = parse_max_values(data)
        self.min_values = parse_min_values(data)
        self.required = parse_required(data)
        return self
    
    
    @copy_docs(ComponentMetadataBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = {}
        
        put_custom_id(self.custom_id, data, defaults)
        put_max_values(self.max_values, data, defaults)
        put_min_values(self.min_values, data, defaults)
        put_required(self.required, data, defaults)
        
        return data
    
    
    @copy_docs(ComponentMetadataBase.clean_copy)
    def clean_copy(self, guild = None):
        new = object.__new__(type(self))
        
        # custom_id
        new.custom_id = self.custom_id
        
        # max_values
        new.max_values = self.max_values
        
        # min_values
        new.min_values = self.min_values
        
        # required
        new.required = self.required
        
        return new
    
    
    @copy_docs(ComponentMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        # custom_id
        new.custom_id = self.custom_id
        
        # max_values
        new.max_values = self.max_values
        
        # min_values
        new.min_values = self.min_values
        
        # required
        new.required = self.required
        
        return new
    
    
    def copy_with(
        self,
        *,
        custom_id = ...,
        max_values = ...,
        min_values = ...,
        required = ...,
    ):
        """
        Copies the attachment input component metadata with the given fields.
        
        Parameters
        ----------
        custom_id : `None | str`, Optional (Keyword only)
            Custom identifier to detect which component was used by the user.
        
        max_values : `int, Optional (Keyword only)
            The maximal amount of attachments to input.
        
        min_values : `int`, Optional (Keyword only)
            The minimal amount of attachments to input.
        
        required : `None | bool`, Optional (Keyword only)
            Whether the field is required to be fulfilled.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # custom_id
        if custom_id is ...:
            custom_id = self.custom_id
        else:
            custom_id = validate_custom_id(custom_id)
        
        # max_values
        if max_values is ...:
            max_values = self.max_values
        else:
            max_values = validate_max_values__attachment_input(max_values)
        
        # min_values
        if min_values is ...:
            min_values = self.min_values
        else:
            min_values = validate_min_values__attachment_input(min_values)
        
        # required
        if required is ...:
            required = self.required
        else:
            required = validate_required(required)
        
        # Extra checks
        
        if custom_id is None:
            custom_id = create_auto_custom_id()
        
        # Construct
        new = object.__new__(type(self))
        new.custom_id = custom_id
        new.max_values = max_values
        new.min_values = min_values
        new.required = required
        return new
    
    
    @copy_docs(ComponentMetadataBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            custom_id = keyword_parameters.pop('custom_id', ...),
            max_values = keyword_parameters.pop('max_values', ...),
            min_values = keyword_parameters.pop('min_values', ...),
            required = keyword_parameters.pop('required', ...),
        )



__all__ = ()

from ...invite import InviteFlag
from ...invite.invite.fields import (
    validate_channel_id, validate_code, validate_flags, validate_inviter_id, validate_max_age, validate_max_uses,
    validate_temporary, validate_uses
)

from ..audit_log_entry_change_conversion import AuditLogEntryChangeConversion, AuditLogEntryChangeConversionGroup
from ..conversion_helpers.converters import value_deserializer_id, value_deserializer_name, value_serializer_id, value_serializer_name


# ---- channel_id ----

CHANNEL_ID_CONVERSION = AuditLogEntryChangeConversion(
    ('channel_id',),
    'channel_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_channel_id,
)


# ---- code ----

CODE_CONVERSION = AuditLogEntryChangeConversion(
    ('code',),
    'code',
    value_deserializer = value_deserializer_name,
    value_serializer = value_serializer_name,
    value_validator = validate_code,
)


# ---- flags ----

FLAGS_CONVERSION = AuditLogEntryChangeConversion(
    ('flags',),
    'flags',
    value_validator = validate_flags,
)


@FLAGS_CONVERSION.set_value_deserializer
def flags_value_deserializer(value):
    if value is None:
        value = InviteFlag()
    else:
        value = InviteFlag(value)
    
    return value


@FLAGS_CONVERSION.set_value_serializer
def flags_value_serializer(value):
    return int(value)


# ---- inviter_id ----

INVITER_ID_CONVERSION = AuditLogEntryChangeConversion(
    ('inviter_id',),
    'inviter_id',
    value_deserializer = value_deserializer_id,
    value_serializer = value_serializer_id,
    value_validator = validate_inviter_id,
)


# ---- max_age ----

MAX_AGE_CONVERSION = AuditLogEntryChangeConversion(
    ('max_age',),
    'max_age',
    value_validator = validate_max_age,
)


# ---- max_uses ----

MAX_USES_CONVERSION = AuditLogEntryChangeConversion(
    ('max_uses',),
    'max_uses',
    value_validator = validate_max_uses,
)


# ---- uses ----

USES_CONVERSION = AuditLogEntryChangeConversion(
    ('uses',),
    'uses',
    value_validator = validate_uses,
)

# ----

@MAX_AGE_CONVERSION.set_value_deserializer
@MAX_USES_CONVERSION.set_value_deserializer
@USES_CONVERSION.set_value_deserializer
def max_value_deserializer(value):
    if value is None:
        value = 0
    return value


# ---- temporary ----

TEMPORARY_CONVERSION = AuditLogEntryChangeConversion(
    ('temporary',),
    'temporary',
    value_validator = validate_temporary,
)


@TEMPORARY_CONVERSION.set_value_deserializer
def temporary_value_deserializer(value):
    if value is None:
        value = False
    return value


# ---- Construct ----

INVITE_CONVERSIONS = AuditLogEntryChangeConversionGroup(
    CHANNEL_ID_CONVERSION,
    CODE_CONVERSION,
    FLAGS_CONVERSION,
    INVITER_ID_CONVERSION,
    MAX_AGE_CONVERSION,
    MAX_USES_CONVERSION,
    TEMPORARY_CONVERSION,
    USES_CONVERSION,
)



__all__ = ()

from ...channel import Channel
from ...field_parsers import bool_parser_factory, int_parser_factory
from ...field_putters import bool_optional_putter_factory, int_putter_factory
from ...field_validators import (
    bool_validator_factory, int_conditional_validator_factory, nullable_entity_array_validator_factory,
    nullable_object_array_validator_factory
)
from ...user import create_partial_user_from_id, thread_user_create

from ..message import Message


def parse_analytics_id(data):
    """
    Parses analytics identifier from teh given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    analytics_id : `int`
    """
    analytics_id = data.get('analytics_id', None)
    if (analytics_id is not None):
        try:
            return int(analytics_id, 16)
        except ValueError:
            pass
    
    return 0


def put_analytics_id(analytics_id, data, defaults):
    """
    Serialises the given analytics identifier.
    
    Parameters
    ----------
    analytics_id : `int`
        The value to serialise.
    
    data : `dict<str, object>`
        Data to serialise into.
    
    defaults : `bool`
        Whether fields as their defaults should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    data['analytics_id'] = format(analytics_id, '0>32x')
    return data


def validate_analytics_id(analytics_id):
    """
    Whether the given `analytics_id` value.
    
    Parameters
    ----------
    analytics_id : `None | int`
        The value to validate.
    
    Returns
    -------
    analytics_id : `int`
    
    Raises
    ------
    TypeError
        - If the given value's type is incorrect.
    ValueError
        - if the given value's value is out of bounds.
    """
    if analytics_id is None:
        return 0
    
    if not isinstance(analytics_id, int):
        raise TypeError(f'`analytics_id` can be `None`, | `int`, got {type(analytics_id).__name__}; {analytics_id!r}.')
    
    if analytics_id < 0:
        raise ValueError(f'`analytics_id` cannot be <= 0, got {analytics_id!r}.')
    
    if analytics_id > (1 << 128) - 1:
        raise ValueError(f'`analytics_id` can be up to 128 bit long; got {analytics_id!r}.')
    
    return analytics_id


# deep_historical_indexing_in_progress

parse_deep_historical_indexing_in_progress = bool_parser_factory('doing_deep_historical_index', False)
put_deep_historical_indexing_in_progress = bool_optional_putter_factory('doing_deep_historical_index', False)
validate_deep_historical_indexing_in_progress = bool_validator_factory('doing_deep_historical_index', False)


# messages


def parse_messages(data):
    """
    Parses the messages out from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    messages : ``None | tuple<Message>``
    """
    message_datas = data.get('messages', None)
    if (message_datas is None) or (not message_datas):
        messages = None
    else:
        messages = (*(Message.from_data(message_data[0]) for message_data in message_datas),)
    
    return messages


def put_messages(messages, data, defaults):
    """
    Serialises the given messages.
    
    Parameters
    ----------
    messages : ``None | tuple<Message>``
        The value to serialise.
    
    data : `dict<str, object>`
        Data to serialise into.
    
    defaults : `bool`
        Whether fields as their default should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if messages is None:
        message_datas = []
    else:
        message_datas = [[message.to_data(defaults = defaults, include_internals = True)] for message in messages]
    
    data['messages'] = message_datas
    return data


validate_messages = nullable_object_array_validator_factory('messages', Message)


# result_count

parse_result_count = int_parser_factory('total_results', 0)
put_result_count = int_putter_factory('total_results')
validate_result_count = int_conditional_validator_factory(
    'result_count',
    0,
    (lambda result_count : result_count >= 0),
    '>= 0',
)


# threads

def parse_threads(data):
    """
    Parses the used threads out from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    threads : ``None | tuple<Channel>``
    """
    channel_datas = data.get('threads', None)
    if (channel_datas is None) or (not channel_datas):
        threads = None
    else:
        threads = tuple(sorted(Channel.from_data(channel_data) for channel_data in channel_datas))
        
        thread_user_datas = data.get('members', None)
        if (thread_user_datas is not None) and thread_user_datas:
            for thread_user_data in thread_user_datas:
                channel_id = int(thread_user_data['id'])
                
                for thread_channel in threads:
                    if thread_channel.id == channel_id:
                        break
                else:
                    continue
                
                user_id = int(thread_user_data['user_id'])
                user = create_partial_user_from_id(user_id)
                thread_user_create(thread_channel, user, thread_user_data)
    
    return threads


def put_threads(threads, data, defaults, *, messages = None):
    """
    Serialises the given threads into the given data.
    
    Parameters
    ----------
    threads : ``None | tuple<Channel>``
        The value to serialise.
    
    data : `dict<str, object>`
        Data to serialise into.
    
    defaults : `bool`
        Whether fields as their default should be included as well.
    
    messages : ``None | tuple<Message>`` = `None`, Optional (Keyword only)
        Messages to detect author guild profiles from.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (threads is not None) or defaults:
        if threads is None:
            channel_datas = []
            thread_user_datas = []
        
        else:
            channel_datas = [channel.to_data(defaults = defaults, include_internals = True) for channel in threads]
            thread_user_datas = []
            
            if (messages is not None):
                for message in messages:
                    channel_id = message.channel_id
                    for thread_channel in threads:
                        if thread_channel.id == channel_id:
                            break
                    else:
                        continue
                    
                    user = message.author
                    thread_profiles = user.thread_profiles
                    if thread_profiles is None:
                        continue
                    
                    try:
                        thread_profile = thread_profiles[channel_id]
                    except KeyError:
                        continue
                    
                    thread_profile_data = thread_profile.to_data(defaults = defaults, include_internals = True)
                    thread_profile_data['id'] = str(channel_id)
                    thread_profile_data['user_id'] = str(user.id)
                    thread_user_datas.append(thread_profile_data)
            
        data['threads'] = channel_datas
        data['members'] = thread_user_datas
    
    return data


validate_threads = nullable_entity_array_validator_factory('channels', Channel)



__all__ = ('message_relative_index',)

from collections import deque
from datetime import datetime as DateTime, timezone as TimeZone
from time import time as current_time

from scarletio import LOOP_TIME, WeakReferer

from ..core import KOKORO
from ..utils import DATETIME_FORMAT_CODE



# searches the relative index of a message in a list
def message_relative_index(messages, message_id):
    """
    Searches the relative index of the given message's id in a channel's message history. The returned index is
    relative, because if the message with the given is not found, it should be at that specific index, if it would be
    inside of the respective channel's message history.
    
    Parameters
    ----------
    messages : `deque` of ``Message``
        The message history of a channel.
    message_id : `int`
        A messages' id to search.
    
    Returns
    -------
    index : `int`
    """
    bot = 0
    top = len(messages)
    while True:
        if bot < top:
            half = (bot + top) >> 1
            if messages[half].id > message_id:
                bot = half + 1
            else:
                top = half
            continue
        break
    
    return bot


class MessageHistoryCollector:
    """
    Attributes
    ----------
    channel_reference : ``WeakReferer`` to ``Channel``
        Reference to the parent channel.
    delay : `float`
        Additional message collection delay.
    handle : `None`, ``TimerHandle``
        Handle calling the history collector.
    """
    __slots__ = ('channel_reference', 'delay', 'handle', )
    
    def __new__(cls, channel, when):
        """
        Creates a new ``MessageHistoryCollector
        """
        self = object.__new__(cls)
        self.channel_reference = WeakReferer(channel)
        self.delay = 0.0
        self.handle = KOKORO.call_at(when, self)
        return self
    
    
    def __call__(self):
        """
        Calls the collector, removing the respective channel's extra messages.
        
        If there is extra delay, moves collection time.
        """
        channel = self.channel_reference()
        if (channel is None):
            self.handle = None
        else:
            delay = self.delay
            if delay:
                self.delay = 0.0
                self.handle = KOKORO.call_at(delay, self)
            else:
                self.handle = None
                channel._message_history_collector = None
                channel._switch_to_limited()
    
    
    def cancel(self):
        """Stops the handler of the collector."""
        handle = self.handle
        if (handle is not None):
            self.handle = None
            handle.cancel()
    
    __del__ = cancel
    
    def add_delay(self, delay):
        """
        Adds delay to the channel history collector.
        
        Parameters
        ----------
        delay : `bool`
            Delay to add.
        """
        self.delay += delay
    
    def __repr__(self):
        """Returns the message history collector's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        handle = self.handle
        if handle is None:
            repr_parts.append(' cancelled')
        else:
            repr_parts.append(' scheduled: ')
            timestamp = DateTime.fromtimestamp(current_time() + handle.when + self.delay - LOOP_TIME(), TimeZone.utc)
            repr_parts.append(timestamp.__format__(DATETIME_FORMAT_CODE))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)


# Do not call any functions from this if you dunno anything about them!
# The message history is basically sorted by message_id, what can be translated to real time.
# The newer messages are at the start, meanwhile the older ones at the end.
# Do not try to delete not existing message's id, or it will cause de-sync.
# Use pypy?

class MessageHistory:
    """
    Contains message logic for message-able channels.
    
    Attributes
    ----------
    _message_keep_limit : `int`
        The channel's own limit of how much messages it should keep before removing their reference.
    _message_history_collector :  `None`, ``MessageHistoryCollector``
        Collector for the channel's message history.
    message_history_reached_end : `bool`
        Whether the channel's message's are loaded till their end. If the channel's message history reach it's end
        no requests will be requested to get older messages.
    messages : `deque` of ``Message`` objects
        The channel's message history.
    """
    __slots__ = ('_message_keep_limit', '_message_history_collector', 'message_history_reached_end', 'messages',)
    
    def __init__(self, message_keep_limit):
        """
        Creates a nwe message history instance with it's default values.
        
        Parameters
        ----------
        message_keep_limit : `int`
            The amount of messages to keep.
        """
        self.message_history_reached_end = False
        self._message_history_collector = None
        self._message_keep_limit = message_keep_limit
        self.messages = None
    
    
    def _set_message_keep_limit(self, message_keep_limit):
        """
        Sets the amount of messages to keep by the channel.
        
        Parameters
        ----------
        message_keep_limit : `int`
            The amount of messages to keep.
        """
        if self._message_keep_limit != message_keep_limit:
            if message_keep_limit == 0:
                new_messages = None
            else:
                old_messages = self.messages
                if old_messages is None:
                    new_messages = None
                else:
                    new_messages = deque(
                        (old_messages[i] for i in range(min(message_keep_limit, len(old_messages)))),
                        maxlen = message_keep_limit,
                    )
            
            self.messages = new_messages
            self._message_keep_limit = message_keep_limit



__all__ = ('Resolver',)

from scarletio import RichAttributeErrorBaseType


class Resolver(RichAttributeErrorBaseType):
    """
    Holds resolver functions for interacting with a resolver.
    
    Attributes
    ----------
    iter_resolve_multiple : ``GeneratorFunctionType``
        Iterable generator for iterating over multiple resolved instances.
    
    iter_resolve_single : ``GeneratorFunctionType``
        Iterable generator for iterating over a single resolved instance.
    
    name : `str`
        The name of the resolver.
    
    resolve_multiple : ``FunctionType``
        Resolver for resolving multiple instances.
    
    resolve_single : ``FunctionType``
        Resolver for resolving a single instance.
    """
    __slots__ = ('iter_resolve_multiple', 'iter_resolve_single', 'name', 'resolve_multiple', 'resolve_single')
    
    def __new__(cls, name, resolve_single, iter_resolve_single, resolve_multiple, iter_resolve_multiple):
        """
        Creates a new resolver instance.
        
        Parameters
        ----------
        name : `str`
            The name of the resolver.
        
        resolve_single : ``FunctionType``
            Resolver for resolving a single instance.
        
        iter_resolve_single : ``GeneratorFunctionType``
            Iterable generator for iterating over a single resolved instance.
        
        resolve_multiple : ``FunctionType``
            Resolver for resolving multiple instances.
        
        iter_resolve_multiple : ``GeneratorFunctionType``
            Iterable generator for iterating over multiple resolved instances.
        """
        self = object.__new__(cls)
        self.iter_resolve_multiple = iter_resolve_multiple
        self.iter_resolve_single = iter_resolve_single
        self.name = name
        self.resolve_multiple = resolve_multiple
        self.resolve_single = resolve_single
        return self
    
    
    def __repr__(self):
        """Returns repr(self)."""
        return f'<{type(self).__name__} {self.name!s}>'



__all__ = ()

from scarletio import copy_docs

from .base import ParameterConverterBase


class ParameterConverterMessageComponentKeyword(ParameterConverterBase):
    """
    Parameter converter for retrieving component value.
    
    Attributes
    ----------
    default : `object`
        Default value of the parameter.
    
    parameter_name : `str`
        The parameter's name.
    """
    __slots__ = ('annotation', 'default')
    
    def __new__(cls, parameter):
        """
        Creates a new parameter converter used by message component fields.
        
        Parameters
        ----------
        parameter : ``Parameter``
            The parameter to create converter from.
        """
        self = object.__new__(cls)
        self.parameter_name = parameter.name
        self.default = parameter.default
        return self
    
    
    @copy_docs(ParameterConverterBase.__call__)
    async def __call__(self, client, interaction_event, value):
        for custom_id, component_type, value_or_values in interaction_event.iter_custom_ids_and_values():
            resolve = component_type.resolve
            if (resolve is not None):
                value_or_values = resolve(interaction_event.resolved, value_or_values)
            break
        
        else:
            value_or_values = None
        
        if value_or_values is None:
            value_or_values = self.default
        
        return value_or_values
    
    
    @copy_docs(ParameterConverterBase._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts):
        # annotation
        repr_parts.append(', annotation = ')
        repr_parts.append(repr(self.annotation))
        
        # default
        repr_parts.append(', default = ')
        repr_parts.append(repr(self.default))



__all__ = ('ApplicationExecutable',)

from scarletio import RichAttributeErrorBaseType

from .fields import (
    parse_launcher, parse_name, parse_os, parse_parameters, put_launcher, put_name, put_os,
    put_parameters, validate_launcher, validate_name, validate_os, validate_parameters
)
from .preinstanced import OperationSystem


class ApplicationExecutable(RichAttributeErrorBaseType):
    """
    Represents a game's executable.
    
    Attributes
    ----------
    launcher : `bool`
        Whether the application is a launcher. Defaults to `False`.
    name : `str`
        The executable's name.
    os : ``OperationSystem``
        The operation system, the executable is for.
    parameters : `None`, `str`
        The parameters to start the application with. Defaults to `None`.
    """
    __slots__ = ('launcher', 'name', 'os', 'parameters')
    
    def __new__(cls, *, launcher = ..., name = ..., os = ..., parameters = ...):
        """
        Creates an application executable.
        
        Parameters
        ----------
        launcher : `bool`, Optional (Keyword only)
            Whether the application is a launcher.
        
        name : `str`, Optional (Keyword only)
            The executable's name.
        
        os : ``OperationSystem``, `str`, Optional (Keyword only)
            The operation system, the executable is for.
        
        parameters : `None`, `str`, Optional (Keyword only)
            The parameters to start the application with.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # launcher
        if launcher is ...:
            launcher = False
        else:
            launcher = validate_launcher(launcher)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # os
        if os is ...:
            os = OperationSystem.none
        else:
            os = validate_os(os)
        
        # parameters
        if parameters is ...:
            parameters = None
        else:
            parameters = validate_parameters(parameters)
        
        self = object.__new__(cls)
        self.launcher = launcher
        self.name = name
        self.os = os
        self.parameters = parameters
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new application executable with the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Executable data.
        """
        self = object.__new__(cls)
        self.launcher = parse_launcher(data)
        self.name = parse_name(data)
        self.os = parse_os(data)
        self.parameters = parse_parameters(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the application executable to json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_launcher(self.launcher, data, defaults)
        put_name(self.name, data, defaults)
        put_os(self.os, data, defaults)
        put_parameters(self.parameters, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the executable's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' ',
            repr(self.name),
        ]
        
        repr_parts.append(', os = ')
        repr_parts.append(repr(self.os))
        
        parameters = self.parameters
        if (parameters is not None):
            repr_parts.append(', parameters = ')
            repr_parts.append(repr(parameters))
        
        if self.launcher:
            repr_parts.append(', launcher = True')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two executables are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # launcher
        if self.launcher != other.launcher:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # os
        if self.os is not other.os:
            return False
        
        # parameters
        if self.parameters != other.parameters:
            return False
        
        return True
    
    
    def __gt__(self, other):
        """Returns whether self is greater than other executable."""
        if type(self) is not type(other):
            return NotImplemented
        
        # name
        self_name = self.name
        other_name = other.name
        if self_name > other_name:
            return True
        
        if self_name < other_name:
            return False
        
        # os
        self_os = self.os
        other_os = other.os
        if self_os > other_os:
            return True
        
        if self_os < other_os:
            return False
        
        # launcher
        self_launcher = self.launcher
        other_launcher = other.launcher
        if self_launcher > other_launcher:
            return True
        
        if self_launcher < other_launcher:
            return False
        
        # parameters
        self_parameters = self.parameters
        other_parameters = other.parameters
        
        if self_parameters is None:
            if (other_parameters is not None):
                return False
        
        else:
            if other_parameters is None:
                return True
            
            if self_parameters > other_parameters:
                return True
        
        return False
    
    
    def __hash__(self):
        """Returns the executable's hash."""
        hash_value = 0
        
        # launcher
        if self.launcher:
            hash_value ^= (1 << 15)
        
        # name
        hash_value ^= hash(self.name)
        
        # os
        hash_value ^= hash(self.os)
        
        # parameters
        parameters = self.parameters
        if (parameters is not None):
            hash_value ^= hash(parameters)
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the application executable.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.launcher = self.launcher
        new.name = self.name
        new.os = self.os
        new.parameters = self.parameters
        return new
    
    
    def copy_with(self, *, launcher = ..., name = ..., os = ..., parameters = ...):
        """
        Copies the application executable with the given fields.
        
        Parameters
        ----------
        launcher : `bool`, Optional (Keyword only)
            Whether the application is a launcher.
        
        name : `str`, Optional (Keyword only)
            The executable's name.
        
        os : ``OperationSystem``, `str`, Optional (Keyword only)
            The operation system, the executable is for.
        
        parameters : `None`, `str`, Optional (Keyword only)
            The parameters to start the application with.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # launcher
        if launcher is ...:
            launcher = self.launcher
        else:
            launcher = validate_launcher(launcher)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # os
        if os is ...:
            os = self.os
        else:
            os = validate_os(os)
        
        # parameters
        if parameters is ...:
            parameters = self.parameters
        else:
            parameters = validate_parameters(parameters)
        
        new = object.__new__(type(self))
        new.launcher = launcher
        new.name = name
        new.os = os
        new.parameters = parameters
        return new



__all__ = ()

from .fields import (
    put_direct_messages_disabled_until, put_invites_disabled_until,
    validate_direct_messages_disabled_duration, validate_direct_messages_disabled_until,
    validate_invites_disabled_duration, validate_invites_disabled_until
)


GUILD_INCIDENTS_FIELD_CONVERTERS = {
    'validate_direct_messages_disabled_duration': (
        validate_direct_messages_disabled_duration, put_direct_messages_disabled_until,
    ),
    'direct_messages_disabled_until': (
        validate_direct_messages_disabled_until, put_direct_messages_disabled_until,
    ),
    'invites_disabled_duration': (validate_invites_disabled_duration, put_invites_disabled_until),
    'invites_disabled_until': (validate_invites_disabled_until, put_invites_disabled_until),
}



__all__ = ()

APPLICATION_ROLE_CONNECTION_KEY = 'guild_connections'
BOOSTER_KEY = 'premium_subscriber'
BOT_ID_KEY = 'bot_id'
INTEGRATION_ID_KEY = 'integration_id'
PURCHASABLE_KEY = 'available_for_purchase'
SUBSCRIPTION_LISTING_ID_KEY = 'subscription_listing_id'



__all__ = ('PollUpdate',)

from itertools import count

from scarletio import RichAttributeErrorBaseType

from ...poll import Poll

from .fields import validate_old_attributes, validate_poll


class PollUpdate(RichAttributeErrorBaseType):
    """
    Represents an updated poll with storing the poll and it's old updated attributes in a `dict`.
    
    Attributes
    ----------
    old_attributes : `dict<str, object>`
        The changed attributes of the poll in `attribute-name` - `old-value` relation. Can contain any of the
        following items:
        
        +---------------------------+-----------------------------------+
        | Keys                      | Values                            |
        +===========================+===================================+
        | allow_multiple_choices    | `bool`                            |
        +---------------------------+-----------------------------------+
        | answers                   | `None`, `tuple` of ``PollAnswer`` |
        +---------------------------+-----------------------------------+
        | duration                  | `int`                             |
        +---------------------------+-----------------------------------+
        | expires_at                | `None | DateTime`                 |
        +---------------------------+-----------------------------------+
        | finalized                 | `bool`                            |
        +---------------------------+-----------------------------------+
        | layout                    | ``PollLayout``                    |
        +---------------------------+-----------------------------------+
        | question                  | `None`, ``PollQuestion``          |
        +---------------------------+-----------------------------------+
    
    poll : ``Poll``
        The updated poll.
    """
    __slots__ = ('poll', 'old_attributes',)
    
    def __new__(cls, *, old_attributes = ..., poll = ...):
        """
        Creates a new poll change instance with the given fields.
        
        Parameters
        ----------
        old_attributes : `None`, `dict<str, object>`, Optional (Keyword only)
            The changed attributes of the poll.
        
        poll : ``Poll``, Optional (Keyword only)
            The updated poll.
        
        Raises
        ------
        TypeError
            - If a field's type is incorrect.
        """
        # old_attributes
        if old_attributes is ...:
            old_attributes = {}
        else:
            old_attributes = validate_old_attributes(old_attributes)
        
        # poll
        if poll is ...:
            poll = Poll()
        else:
            poll = validate_poll(poll)
        
        # Construct
        self = object.__new__(cls)
        self.old_attributes = old_attributes
        self.poll = poll
        return self
    
    
    @classmethod
    def from_fields(cls, poll, old_attributes):
        """
        Creates a new poll change instance with the given fields.
        
        Parameters
        ----------
        poll : ``Poll``
            The updated poll.
        
        old_attributes : `dict<str, object>`
            The changed attributes of the poll.
        
        Returns
        -------
        new : `instance<cls>`
        """
        self = object.__new__(cls)
        self.old_attributes = old_attributes
        self.poll = poll
        return self
    
    
    def __repr__(self):
        """Returns the representation of the poll update."""
        repr_parts = ['<', type(self).__name__]
        
        repr_parts.append(' poll = ')
        repr_parts.append(repr(self.poll))
        
        repr_parts.append(', old_attributes = ')
        repr_parts.append(repr(self.old_attributes))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the poll update's hash value."""
        hash_value = 0
        
        # old_attributes
        old_attributes = self.old_attributes
        hash_value ^= len(old_attributes)
        
        for mask, key in zip(count(5, 7), sorted(old_attributes.keys())):
            hash_value ^= (mask | hash(key)) & hash(old_attributes[key])
        
        # poll
        hash_value ^= hash(self.poll)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two poll updates are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.old_attributes != other.old_attributes:
            return False
        
        if self.poll != other.poll:
            return False
        
        return True
    
    
    def __len__(self):
        """Helper for unpacking if needed."""
        return 2
    
    
    def __iter__(self):
        """
        Unpacks the poll update.
        
        This method is a generator.
        """
        yield self.poll
        yield self.old_attributes
    
    
    def copy(self):
        """
        Copies the poll update.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.old_attributes = self.old_attributes.copy()
        new.poll = self.poll.copy()
        return new
    
    
    def copy_with(self, *, old_attributes = ..., poll = ...):
        """
        Copies the poll change with the given fields.
        
        Parameters
        ----------
        old_attributes : `None`, `dict<str, object>`, Optional (Keyword only)
            The changed attributes of the poll.
        
        poll : ``Poll``, Optional (Keyword only)
            The updated poll.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a field's type is incorrect.
        """
        # old_attributes
        if old_attributes is ...:
            old_attributes = self.old_attributes.copy()
        else:
            old_attributes = validate_old_attributes(old_attributes)
        
        # poll
        if poll is ...:
            poll = self.poll.copy()
        else:
            poll = validate_poll(poll)
        
        # Construct
        new = object.__new__(type(self))
        new.poll = poll
        new.old_attributes = old_attributes
        return new



__all__ = ()


def merge_update_poll_results(new_results, old_results):
    """
    Merges poll results.
    
    Parameters
    ----------
    new_results : `None | list<PollResult>`
        New results.
    old_results : `None | list<PollResult>`
        Old results.
    
    Returns
    -------
    results : `None | list<PollResult>`
    """
    if new_results is None:
        return None
    
    if old_results is None:
        return new_results
    
    # Collect alive answer_id-s
    answer_ids = {result.answer_id for result in new_results}
    
    # Remove not alive answer_id-s
    for index in reversed(range(len(old_results))):
        old_result = old_results[index]
        if old_result.answer_id not in answer_ids:
            del old_results[index]
    
    # Merge old with new.
    for new_result in new_results:
        answer_id = new_result.answer_id
        for old_result in old_results:
            if answer_id == old_result.answer_id:
                old_result._merge_with(new_result)
                break
        else:
            old_results.append(new_result)

    return old_results


def poll_result_sort_key(result):
    """
    Sort key that can be used for poll results.
    
    Parameters
    ----------
    result : ``PollResult``
        Poll Result to get its sort key of.
    
    Returns
    -------
    sort_key : `int`
    """
    return result.answer_id



__all__ = ()

from ...field_parsers import default_date_time_parser_factory, nullable_entity_parser_factory
from ...field_putters import force_date_time_putter_factory
from ...field_validators import force_date_time_validator_factory, nullable_entity_validator_factory
from ...utils import DISCORD_EPOCH_START

from ..message import Message


# message

parse_message = nullable_entity_parser_factory('message', Message)

def put_message(message, data, defaults):
    """
    Puts the given `message` into the given data.
    
    Parameters
    ----------
    message : ``None | Message``
        The message to put into the given `data`.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default fields should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (message is not None):
        if message is None:
            entity_data = None
        else:
            entity_data = message.to_data(defaults = defaults, include_internals = True)
        
        data['message'] = entity_data
    
    return data

validate_message = nullable_entity_validator_factory('message', Message)


# pinned_at

parse_pinned_at = default_date_time_parser_factory('pinned_at', DISCORD_EPOCH_START)
put_pinned_at = force_date_time_putter_factory('pinned_at')
validate_pinned_at = force_date_time_validator_factory('pinned_at')



__all__ = ('VoiceChannelEffect',)

from scarletio import copy_docs

from ...bases import EventBase
from ...core import GUILDS
from ...soundboard import create_partial_soundboard_sound_from_id
from ...user import create_partial_user_from_id

from ..channel import ChannelType, create_partial_channel_from_id

from .fields import (
    parse_animation_id, parse_animation_type, parse_channel_id, parse_emoji, parse_guild_id, parse_user_id,
    put_animation_id, put_animation_type, put_channel_id, put_emoji, put_guild_id,
    put_user_id, validate_animation_id, validate_animation_type, validate_channel_id, validate_emoji,
    validate_guild_id, validate_user_id, parse_sound_id, put_sound_id, validate_sound_id,
    parse_sound_volume, put_sound_volume, validate_sound_volume
)
from .preinstanced import VoiceChannelEffectAnimationType


class VoiceChannelEffect(EventBase):
    """
    Represents a voice channel effect or a played sound.
    
    Attributes
    ----------
    animation_id : `int`
        The identifier of the animation of the voice channel effect.
    animation_type : ``VoiceChannelEffectAnimationType``
        The animation's type.
    channel_id : `int`
        The channel's identifier where the effect was sent.
    emoji : ``None | Emoji``
        The emoji sent.
    guild_id : `int`
        The guild's identifier where the event was sent.
    sound_id : `int`
        The played sound's identifier.
    sound_volume : `float`
        The played sound's volume.
    user_id : `int`
        The user's identifier who sent the effect.
    """
    __slots__ = (
        'animation_id', 'animation_type', 'channel_id', 'emoji', 'guild_id', 'sound_id', 'sound_volume', 'user_id'
    )
    
    def __new__(
        cls,
        *,
        animation_id = ...,
        animation_type = ...,
        channel_id = ...,
        emoji = ...,
        guild_id = ...,
        sound_id = ...,
        sound_volume = ...,
        user_id = ...,
    ):
        """
        Creates a new voice channel effect event.
        
        Parameters
        ----------
        animation_id : `int`, Optional (Keyword only)
            The identifier of the animation of the voice channel effect.
        animation_type : ``VoiceChannelEffectAnimationType``, `int`, Optional (Keyword only)
            The animation's type.
        channel_id : `int`, ``Channel``, Optional (Keyword only)
            The channel or its identifier where the effect was sent.
        emoji : ``None | Emoji``, Optional (Keyword only)
            The emoji sent.
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The guild or its identifier where the event was sent.
        sound_id : `int`, ``SoundboardSound``, Optional (Keyword only)
            The played sound's identifier.
        sound_volume : `float`, Optional (Keyword only)
            The played sound's volume.
        user_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The user or its identifier who sent the effect.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # animation_id
        if animation_id is ...:
            animation_id = 0
        else:
            animation_id = validate_animation_id(animation_id)
        
        # animation_type
        if animation_type is ...:
            animation_type = VoiceChannelEffectAnimationType.premium
        else:
            animation_type = validate_animation_type(animation_type)
        
        # channel_id
        if channel_id is ...:
            channel_id = 0
        else:
            channel_id = validate_channel_id(channel_id)
        
        # emoji
        if emoji is ...:
            emoji = None
        else:
            emoji = validate_emoji(emoji)
        
        # guild_id
        if guild_id is ...:
            guild_id = 0
        else:
            guild_id = validate_guild_id(guild_id)
        
        # sound_id
        if sound_id is ...:
            sound_id = 0
        else:
            sound_id = validate_sound_id(sound_id)
        
        # sound_volume
        if sound_volume is ...:
            sound_volume = 1.0
        else:
            sound_volume = validate_sound_volume(sound_volume)
        
        # user_id
        if user_id is ...:
            user_id = 0
        else:
            user_id = validate_user_id(user_id)
        
        # Construct
        self = object.__new__(cls)
        self.animation_id = animation_id
        self.animation_type = animation_type
        self.channel_id = channel_id
        self.emoji = emoji
        self.guild_id = guild_id
        self.sound_id = sound_id
        self.sound_volume = sound_volume
        self.user_id = user_id
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a voice channel effect event from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Voice channel effect event data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.animation_id = parse_animation_id(data)
        self.animation_type = parse_animation_type(data)
        self.channel_id = parse_channel_id(data)
        self.emoji = parse_emoji(data)
        self.guild_id = parse_guild_id(data)
        self.sound_id = parse_sound_id(data)
        self.sound_volume = parse_sound_volume(data)
        self.user_id = parse_user_id(data)
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the voice channel effect event into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_animation_id(self.animation_id, data, defaults)
        put_animation_type(self.animation_type, data, defaults)
        put_emoji(self.emoji, data, defaults)
        put_sound_id(self.sound_id, data, defaults)
        put_sound_volume(self.sound_volume, data, defaults)
        
        if include_internals:
            put_channel_id(self.channel_id, data, defaults)
            put_guild_id(self.guild_id, data, defaults)
            put_user_id(self.user_id, data, defaults)
        
        return data
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' guild_id = ')
        repr_parts.append(repr(self.guild_id))
        
        repr_parts.append(', channel_id = ')
        repr_parts.append(repr(self.channel_id))
        
        repr_parts.append(', user_id = ')
        repr_parts.append(repr(self.user_id))
        
        emoji = self.emoji
        if (emoji is not None):
            repr_parts.append(', emoji = ')
            repr_parts.append(repr(emoji))
            
            animation_type = self.animation_type
            repr_parts.append(', animation_type = ')
            repr_parts.append(animation_type.name)
            repr_parts.append(' ~ ')
            repr_parts.append(repr(animation_type.value))
            
            repr_parts.append(', animation_id = ')
            repr_parts.append(repr(self.animation_id))
        
        sound_id = self.sound_id
        if sound_id:
            repr_parts.append(', sound_id = ')
            repr_parts.append(repr(sound_id))
            
            repr_parts.append(', sound_volume = ')
            repr_parts.append(repr(self.sound_volume))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 0
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        return
        yield
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # animation_id
        if self.animation_id != other.animation_id:
            return False
        
        # animation_type
        if self.animation_type is not other.animation_type:
            return False
        
        # channel_id
        if self.channel_id is not other.channel_id:
            return False
        
        # emoji
        if self.emoji is not other.emoji:
            return False
        
        # guild_id
        if self.guild_id != other.guild_id:
            return False
        
        # sound_id
        if self.sound_id != other.sound_id:
            return False
        
        # sound_volume
        if self.sound_volume != other.sound_volume:
            return False
        
        # user_id
        if self.user_id != other.user_id:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # animation_id
        hash_value ^= hash(self.animation_id)
        
        # animation_type
        hash_value ^= hash(self.animation_type)
        
        # channel_id
        hash_value ^= self.channel_id
        
        # emoji
        emoji = self.emoji
        if (emoji is not None):
            hash_value ^= hash(emoji)
        
        # guild_id
        hash_value ^= self.guild_id
        
        # sound_id
        hash_value ^= self.sound_id
        
        # sound_volume
        sound_volume = self.sound_volume
        if (sound_volume != 1.0):
            hash_value ^= hash(sound_volume)
        
        # user_id
        hash_value ^= self.user_id
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the voice channel effect event.
        
        Returns
        -------
        new : `instance<type<new>>`
        """
        new = object.__new__(type(self))
        new.animation_id = self.animation_id
        new.animation_type = self.animation_type
        new.channel_id = self.channel_id
        new.emoji = self.emoji
        new.guild_id = self.guild_id
        new.sound_id = self.sound_id
        new.sound_volume = self.sound_volume
        new.user_id = self.user_id
        return new
    
    
    def copy_with(
        self,
        *,
        animation_id = ...,
        animation_type = ...,
        channel_id = ...,
        emoji = ...,
        guild_id = ...,
        sound_id = ...,
        sound_volume = ...,
        user_id = ...,
    ):
        """
        Copies new voice channel effect event with the given fields.
        
        Parameters
        ----------
        animation_id : `int`, Optional (Keyword only)
            The identifier of the animation of the voice channel effect.
        animation_type : ``VoiceChannelEffectAnimationType``, `int`, Optional (Keyword only)
            The animation's type.
        channel_id : `int`, ``Channel``, Optional (Keyword only)
            The channel or its identifier where the effect was sent.
        emoji : ``None | Emoji``, Optional (Keyword only)
            The emoji sent.
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The guild or its identifier where the event was sent.
        sound_id : `int`, ``SoundboardSound``, Optional (Keyword only)
            The played sound's identifier.
        sound_volume : `float`, Optional (Keyword only)
            The played sound's volume.
        user_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The user or its identifier who sent the effect.
        
        Returns
        -------
        new : `instance<type<new>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # animation_id
        if animation_id is ...:
            animation_id = self.animation_id
        else:
            animation_id = validate_animation_id(animation_id)
        
        # animation_type
        if animation_type is ...:
            animation_type = self.animation_type
        else:
            animation_type = validate_animation_type(animation_type)
        
        # channel_id
        if channel_id is ...:
            channel_id = self.channel_id
        else:
            channel_id = validate_channel_id(channel_id)
        
        # emoji
        if emoji is ...:
            emoji = self.emoji
        else:
            emoji = validate_emoji(emoji)
        
        # guild_id
        if guild_id is ...:
            guild_id = self.guild_id
        else:
            guild_id = validate_guild_id(guild_id)
        
        # sound_id
        if sound_id is ...:
            sound_id = self.sound_id
        else:
            sound_id = validate_sound_id(sound_id)
        
        # sound_volume
        if sound_volume is ...:
            sound_volume = self.sound_volume
        else:
            sound_volume = validate_sound_volume(sound_volume)
        
        # user_id
        if user_id is ...:
            user_id = self.user_id
        else:
            user_id = validate_user_id(user_id)
        
        # Construct
        new = object.__new__(type(self))
        new.animation_id = animation_id
        new.animation_type = animation_type
        new.channel_id = channel_id
        new.emoji = emoji
        new.guild_id = guild_id
        new.sound_id = sound_id
        new.sound_volume = sound_volume
        new.user_id = user_id
        return new
    
    
    @property
    def channel(self):
        """
        Returns the voice channel effect event's channel.
        
        Returns
        -------
        channel : ``Channel``
        """
        return create_partial_channel_from_id(self.channel_id, ChannelType.guild_voice, self.guild_id)
    
    
    @property
    def guild(self):
        """
        Returns the voice channel effect's guild. If the guild is not cached returns `None`.
        
        Returns
        -------
        guild : ``None | Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    @property
    def user(self):
        """
        Returns the voice channel effect event's user.
        
        Returns
        -------
        user : ``ClientUserBase``
        """
        return create_partial_user_from_id(self.user_id)
    
    
    @property
    def sound(self):
        """
        Returns the voice channel effect event's sound.
        
        Returns
        -------
        sound : ``None | SoundboardSound``
        """
        sound_id = self.sound_id
        if sound_id:
            return create_partial_soundboard_sound_from_id(sound_id, self.guild_id)



__all__ = ()

from ...exceptions import DiscordException, ERROR_CODES


async def _delete_reaction_with_task(reaction_add_event, client):
    """
    Removes the given reaction event's emoji from it's message if applicable. Expected error codes are silenced.
    
    Parameters
    ----------
    reaction_add_event : ``ReactionAddEvent``, ``ReactionDeleteEvent``
        The respective reaction event.
    client : ``Client``
        The client who should remove the reaction if applicable.
    """
    try:
        await client.reaction_delete(reaction_add_event.message, reaction_add_event.reaction, reaction_add_event.user)
    except GeneratorExit:
        raise
    
    except BaseException as err:
        
        if isinstance(err, ConnectionError):
            # no internet
            return
        
        if isinstance(err, DiscordException):
            if err.code in (
                ERROR_CODES.unknown_message, # message deleted
                ERROR_CODES.unknown_channel, # channel deleted
                ERROR_CODES.missing_access, # client removed
                ERROR_CODES.missing_permissions, # permissions changed meanwhile
            ):
                return
        
        await client.events.error(client, f'_delete_reaction_with_task called from {reaction_add_event!r}', err)
        return



__all__ = ('WebhookSourceGuild', )

from scarletio import include

from ...bases import DiscordEntity, IconSlot
from ...http.urls import build_guild_icon_url, build_guild_icon_url_as

from .fields import validate_id, validate_name, parse_id, parse_name, put_id, put_name


create_partial_guild_from_id = include('create_partial_guild_from_id')


class WebhookSourceGuild(DiscordEntity):
    """
    Entity representing a server type ``Webhook``'s guild.
    
    Attributes
    ----------
    id : `int`
        The unique identifier number of the guild.
    name : `str`
        The name of the guild.
    icon_hash : `int`
        The guild's icon's hash in `uint128`.
    icon_type : ``IconType``
        The guild's icon's type.
    """
    __slots__ = ('name',)
    
    icon = IconSlot('icon', 'icon')
    
    def __new__(cls, *, guild_id = ..., icon = ..., name = ...):
        """
        Creates a new webhook source guild from the given fields.
        
        Parameters
        ----------
        guild_id : `int`, Optional (Keyword only)
            The guild's identifier.
        icon : ``None | str | Icon``, Optional (Keyword only)
            The guild's icon.
        name : `str`, Optional (Keyword only)
            The guild's name.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        """
        # guild_id
        if guild_id is ...:
            guild_id = 0
        else:
            guild_id = validate_id(guild_id)
        
        # icon
        if icon is ...:
            icon = None
        else:
            icon = cls.icon.validate_icon(icon)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # Constructor
        self = object.__new__(cls)
        self.icon = icon
        self.id = guild_id
        self.name = name
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new webhook source guild from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Webhook source guild data.
        
        Returns
        -------
        new : `instance<cls>`
        """
        self = object.__new__(cls)
        self._set_icon(data)
        self.id = parse_id(data)
        self.name = parse_name(data)
        
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the webhook source guild to json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        type(self).icon.put_into(self.icon, data, defaults)
        put_id(self.id, data, defaults)
        put_name(self.name, data, defaults)
        return data
    
    
    @classmethod
    def from_guild(cls, guild):
        """
        Creates a new webhook source guild from the given guild.
        
        Parameters
        ----------
        guild : ``Guild``
            The respective guild instance.
        
        Returns
        -------
        self : ``WebhookSourceGuild``
        """
        self = object.__new__(cls)
        self.icon_hash = guild.icon_hash
        self.icon_type = guild.icon_type
        self.id = guild.id
        self.name = guild.name
        return self
    
    
    def __repr__(self):
        """Returns the webhook source guild's representation."""
        return f'<{self.__class__.__name__} id = {self.id!r}, name = {self.name!r}>'
    
    
    def __hash__(self):
        """Returns the webhook guild source's hash value."""
        hash_value = 0
        
        # icon
        hash_value ^= hash(self.icon)
        
        # id
        hash_value ^= self.id
        
        # name
        hash_value ^= hash(self.name)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two webhook source guilds are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two webhook source guilds are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether self equals to other. `other` must be same type as self.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other webhook source guild to compare self to.
        
        Returns
        -------
        is_equal : `bool`
        """
        # icon
        if self.icon_hash != other.icon_hash:
            return False
        if self.icon_type != other.icon_type:
            return False
        
        # id
        if self.id != other.id:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        return True
    
    
    @property
    def guild(self):
        """
        Returns the guild of the webhook source guild.
        
        If the guild is not cached, creates a new one.
        
        Returns
        -------
        guild : ``Guild``
        """
        guild = create_partial_guild_from_id(self.id)
        if guild.partial:
            guild.name = self.name
            guild.icon_hash = self.icon_hash
            guild.icon_type = self.icon_type
        
        return guild
    
    
    def copy(self):
        """
        Copies the webhook source guild.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.icon = self.icon
        new.id = self.id
        new.name = self.name
        return new
    
    
    def copy_with(self, *, guild_id = ..., icon = ..., name = ...):
        """
        Copies a new webhook source guild from the given fields.
        
        Parameters
        ----------
        guild_id : `int`, Optional (Keyword only)
            The guild's identifier.
        icon : ``None | str | Icon``, Optional (Keyword only)
            The guild's icon.
        name : `str`, Optional (Keyword only)
            The guild's name.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        """
        # guild_id
        if guild_id is ...:
            guild_id = self.id
        else:
            guild_id = validate_id(guild_id)
        
        # icon
        if icon is ...:
            icon = self.icon
        else:
            icon = type(self).icon.validate_icon(icon)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # Constructor
        new = object.__new__(type(self))
        new.icon = icon
        new.id = guild_id
        new.name = name
        return new
    
    
    @property
    def icon_url(self):
        """
        Returns the guild's icon's url. If the guild has no icon, then returns `None`.
        
        Returns
        -------
        url : `None | str`
        """
        return build_guild_icon_url(self.id, self.icon_type, self.icon_hash)
    
    
    def icon_url_as(self, ext = None, size = None):
        """
        Returns the guild's icon's url. If the guild has no icon, then returns `None`.
        
        Parameters
        ----------
        ext : `None | str` = `None`, Optional
            The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
            If the guild has animated icon, it can be `'gif'` as well.
        
        size : `None | int` = `None`, Optional
            The preferred minimal size of the image's url.
        
        Returns
        -------
        url : `None | str`
        """
        return build_guild_icon_url_as(self.id, self.icon_type, self.icon_hash, ext, size)



__all__ = (
    'API_ENDPOINT', 'CDN_ENDPOINT', 'DISCORD_ENDPOINT', 'INVITE_URL_RP', 'STATUS_ENDPOINT', 'VALID_ICON_FORMATS',
    'VALID_ICON_FORMATS_EXTENDED', 'VALID_ICON_MEDIA_TYPES', 'VALID_ICON_MEDIA_TYPES_EXTENDED',
    'VALID_IMAGE_MEDIA_TYPES_ALL', 'VALID_STICKER_IMAGE_MEDIA_TYPES', 'is_media_url', 'parse_message_jump_url'
)

from itertools import chain
from re import compile as re_compile

from scarletio import export, include

from ...env import (
    API_VERSION, CUSTOM_API_ENDPOINT, CUSTOM_CDN_ENDPOINT, CUSTOM_DISCORD_ENDPOINT, CUSTOM_INVITE_ENDPOINT,
    CUSTOM_MEDIA_ENDPOINT, CUSTOM_STATUS_ENDPOINT
)


StickerFormat = include('StickerFormat')
_try_get_guild_id = include('_try_get_guild_id')


API_ENDPOINT = f'https://discord.com/api/v{API_VERSION}' if (CUSTOM_API_ENDPOINT is None) else CUSTOM_API_ENDPOINT
CDN_ENDPOINT = 'https://cdn.discordapp.com' if (CUSTOM_CDN_ENDPOINT is None) else CUSTOM_CDN_ENDPOINT
DISCORD_ENDPOINT = 'https://discord.com' if (CUSTOM_DISCORD_ENDPOINT is None) else CUSTOM_DISCORD_ENDPOINT
STATUS_ENDPOINT = 'https://status.discord.com/api/v2' if (CUSTOM_STATUS_ENDPOINT is None) else CUSTOM_STATUS_ENDPOINT
MEDIA_ENDPOINT =  'https://media.discordapp.net' if (CUSTOM_MEDIA_ENDPOINT is None) else CUSTOM_MEDIA_ENDPOINT
INVITE_ENDPOINT = 'https://discord.gg' if (CUSTOM_INVITE_ENDPOINT is None) else CUSTOM_INVITE_ENDPOINT

del CUSTOM_API_ENDPOINT, CUSTOM_CDN_ENDPOINT, CUSTOM_DISCORD_ENDPOINT, CUSTOM_STATUS_ENDPOINT, API_VERSION

VALID_ICON_SIZES = frozenset((
    *( 1 << x       for x in       range(4, 13)               ),
    *((1 << x) *  3 for x in chain(range(3,  6), range(9, 11))),
    *((1 << x) *  5 for x in       range(2,  9)               ),
    *((1 << x) *  7 for x in       range(2,  4)               ),
    *((1 << x) * 11 for x in       range(1,  3)               ),
))

VALID_ICON_FORMATS = frozenset(('jpg', 'jpeg', 'png', 'webp'))
VALID_ICON_FORMATS_EXTENDED = frozenset((*VALID_ICON_FORMATS, 'gif',))

VALID_ICON_MEDIA_TYPES = frozenset(('image/jpeg', 'image/png', 'image/webp'))
VALID_ICON_MEDIA_TYPES_EXTENDED = frozenset(('image/gif', *VALID_ICON_MEDIA_TYPES))

VALID_STICKER_IMAGE_MEDIA_TYPES = frozenset(('image/gif', 'image/png', 'application/json'))
VALID_IMAGE_MEDIA_TYPES_ALL = frozenset((*VALID_ICON_MEDIA_TYPES_EXTENDED, *VALID_STICKER_IMAGE_MEDIA_TYPES))


WIDGET_STYLE_RP = re_compile('shield|banner[1-4]')

MESSAGE_JUMP_URL_RP = re_compile(
    '(?:https://)?(?:(?:canary|ptb)\\.)?discord(?:app)?.com/channels/(?:(\\d{7,21})|@me)/(\\d{7,21})/(\\d{7,21})'
)
export(MESSAGE_JUMP_URL_RP, 'MESSAGE_JUMP_URL_RP')


CDN_RP = re_compile(
    'https://(?:'
        'cdn\\.discordapp\\.com|'
        '(?:(?:canary|ptb)\\.)?discord\\.com|'
        '(?:'
            'images-ext-\\d+|'
            'media'
        ')\\.discordapp\\.net'
    ')/'
)


WEBHOOK_URL_PATTERN = re_compile(
    '(?:https://)?discord(?:app)?.com/api/(?:v\\d+/)?webhooks/([0-9]{17,21})/([a-zA-Z0-9.\\-_%]{60,68})(?:/.*)?'
)

INVITE_URL_RP = re_compile('(?:https?://)?discord(?:\\.gg|(?:app)?\\.com/invite)/([a-zA-Z0-9-]+)')


def _validate_extension(icon_type, ext):
    """
    Validates the given icon extension.
    
    Parameters
    ----------
    icon_type : ``IconType``
        The respective icon type.
    
    ext : `None | str`
        The received extension.
    
    Returns
    -------
    ext : `str`
        The validated extension.
    
    Raises
    ------
    ValueError
        - If `ext`'s value is not applicable for the given icon type.
    """
    if ext is None:
        ext = icon_type.default_postfix
    
    else:
        if not icon_type.allows_postfix(ext):
            raise ValueError(
                f'Extension must be one of {icon_type.allowed_postfixes}, got {ext!r}.'
            )
    
    return ext


def _build_end(size, add_animated_query_parameter):
    """
    Validates the given icon size.
    
    Parameters
    ----------
    size : `None | int`
        The received size.
    
    add_animated_query_parameter : `bool`
        Whether `animated=true` query parameter should be added.
    
    Returns
    -------
    end : `str`
        The validated size as query string.
    
    Raises
    ------
    ValueError
        - If `size`'s value is not applicable for the given icon type.
    """
    if size is None:
        end = ''
    
    elif size in VALID_ICON_SIZES:
        end = f'?size={size}'
    
    else:
        raise ValueError(
            f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size!r}.'
        )
    
    if add_animated_query_parameter:
        if end:
            end += '&animated=true'
        else:
            end = '?animated=true'
    
    return end


def build_activity_asset_image_invite_cover_url(application_id, image_invite_cover):
    """
    Returns the activity's invite cover asset image's url.
    If the activity has no invite cover asset image, then returns `None`.
    
    Parameters
    ----------
    application_id : `int`
        The activity's application's identifier.
    
    image_invite_cover : `None | str`
        The activity's asset's invite cover image's value.
    
    Returns
    -------
    url : `None | str`
    """
    if (not application_id) or (image_invite_cover is None) or (not image_invite_cover.isdigit()):
        return None
    
    return f'{CDN_ENDPOINT}/app-assets/{application_id}/{image_invite_cover}.png'


def build_activity_asset_image_invite_cover_url_as(application_id, image_invite_cover, ext, size):
    """
    Returns the activity's invite cover asset image's url.
    If the activity has no invite cover asset image, then returns `None`.
    
    Parameters
    ----------
    application_id : `int`
        The activity's application's identifier.
    
    image_invite_cover : `None | str`
        The activity's asset's invite cover image's value.
    
    ext : `None | str`
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    
    size : `None | int`
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    if (not application_id) or (image_invite_cover is None) or (not image_invite_cover.isdigit()):
        return None
    
    end = _build_end(size, False)

    if ext is None:
        ext = 'png'
    elif ext not in VALID_ICON_FORMATS:
        raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')
    
    return f'{CDN_ENDPOINT}/app-assets/{application_id}/{image_invite_cover}.{ext}{end}'



def build_activity_asset_image_large_url(application_id, image_large):
    """
    Returns the activity's large asset image's url. If the activity has no large asset image, then returns `None`.
    
    Parameters
    ----------
    application_id : `int`
        The activity's application's identifier.
    
    image_large : `None | str`
        The activity's asset's large image's value.
    
    Returns
    -------
    url : `None | str`
    """
    if (not application_id) or (image_large is None) or (not image_large.isdigit()):
        return None
    
    return f'{CDN_ENDPOINT}/app-assets/{application_id}/{image_large}.png'


def build_activity_asset_image_large_url_as(application_id, image_large, ext, size):
    """
    Returns the activity's large asset image's url. If the activity has no large asset image, then returns `None`.
    
    Parameters
    ----------
    application_id : `int`
        The activity's application's identifier.
    
    image_large : `None | str`
        The activity's asset's large image's value.
    
    ext : `None | str`
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    
    size : `None | int`
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    if (not application_id) or (image_large is None) or (not image_large.isdigit()):
        return None
    
    end = _build_end(size, False)

    if ext is None:
        ext = 'png'
    elif ext not in VALID_ICON_FORMATS:
        raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')
    
    return f'{CDN_ENDPOINT}/app-assets/{application_id}/{image_large}.{ext}{end}'


def build_activity_asset_image_small_url(application_id, image_small):
    """
    Returns the activity's small asset image's url. If the activity has no small asset image, then returns `None`.
    
    Parameters
    ----------
    application_id : `int`
        The activity's application's identifier.
    
    image_small : `None | str`
        The activity's asset's small image's value.
    
    Returns
    -------
    url : `None | str`
    """
    if (not application_id) or (image_small is None) or (not image_small.isdigit()):
        return None
    
    return f'{CDN_ENDPOINT}/app-assets/{application_id}/{image_small}.png'


def build_activity_asset_image_small_url_as(application_id, image_small, ext, size):
    """
    Returns the activity's small asset image's url. If the activity has no small asset image, then returns `None`.
    
    Parameters
    ----------
    application_id : `int`
        The activity's application's identifier.
    
    image_small : `None | str`
        The activity's asset's small image's value.
    
    ext : `None | str`
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    
    size : `None | int`
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    if (not application_id) or (image_small is None) or (not image_small.isdigit()):
        return None
    
    end = _build_end(size, False)

    if ext is None:
        ext = 'png'
    elif ext not in VALID_ICON_FORMATS:
        raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')
    
    return f'{CDN_ENDPOINT}/app-assets/{application_id}/{image_small}.{ext}{end}'


def build_application_cover_url(application_id, icon_type, icon_hash):
    """
    Returns the application's cover image's url. If the application has no cover image, then returns `None`.
    
    Parameters
    ----------
    application_id : `int`
        The application's identifier.
    
    icon_type : ``IconType``
        The application's cover icon's type.
    
    icon_hash : `int`
        The application's cover icon's hash (uint128).
    
    Returns
    -------
    url : `None | str`
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/app-assets/{application_id}/store/{prefix}{icon_hash:0>32x}.{ext}'


def build_application_cover_url_as(application_id, icon_type, icon_hash, ext, size):
    """
    Returns the application's cover image's url. If the application has no cover image, then returns `None`.
    
    Parameters
    ----------
    application_id : `int`
        The application's identifier.
    
    icon_type : ``IconType``
        The application's cover icon's type.
    
    icon_hash : `int`
        The application's cover icon's hash (uint128).
    
    ext : `None | str`
        The extension of the cover's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    
    size : `None | int`
        The preferred minimal size of the cover's url.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size, prefix == 'a_' and ext == 'webp')
    
    return f'{CDN_ENDPOINT}/app-assets/{application_id}/store/{prefix}{icon_hash:0>32x}.{ext}{end}'


def build_application_icon_url(application_id, icon_type, icon_hash):
    """
    Returns the application's icon's url. If the application has no icon, then returns `None`.
    
    Parameters
    ----------
    application_id : `int`
        The application's identifier.
    
    icon_type : ``IconType``
        The application's icon's type.
    
    icon_hash : `int`
        The application's icon's hash (uint128).
    
    Returns
    -------
    url : `None | str`
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/app-icons/{application_id}/{prefix}{icon_hash:0>32x}.{ext}'


def build_application_icon_url_as(application_id, icon_type, icon_hash, ext, size):
    """
    Returns the application's icon's url. If the application has no icon, then returns `None`.
    
    Parameters
    ----------
    application_id : `int`
        The application's identifier.
    
    icon_type : ``IconType``
        The application's icon's type.
    
    icon_hash : `int`
        The application's icon's hash (uint128).
    
    ext : `None | str`
        The extension of the icon's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    
    size : `None | int`
        The preferred minimal size of the icon's url.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size, prefix == 'a_' and ext == 'webp')
    
    return f'{CDN_ENDPOINT}/app-icons/{application_id}/{prefix}{icon_hash:0>32x}.{ext}{end}'


def build_avatar_decoration_url(icon_type, icon_hash):
    """
    Returns an avatar decoration's url. If the avatar decoration has no url, returns `None`.
    
    Parameters
    ----------
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    Returns
    -------
    url : `None | str`
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/avatar-decoration-presets/{prefix}{icon_hash:0>32x}.{ext}'


def build_avatar_decoration_url_as(icon_type, icon_hash, ext, size):
    """
    Returns an avatar decoration's url. If the avatar decoration has no url, returns `None`.
    
    Parameters
    ----------
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    ext : `None | str`
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    
    size : `None | int`
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size, prefix == 'a_' and ext == 'webp')
    
    return f'{CDN_ENDPOINT}/avatar-decoration-presets/{prefix}{icon_hash:0>32x}.{ext}{end}'


def build_channel_group_icon_url(channel_id, icon_type, icon_hash):
    """
    Returns the group channel's icon's image's url. If the channel has no icon, then returns `None`.
    
    Parameters
    ----------
    channel_id : `int`
        The channel's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    Returns
    -------
    url : `None | str`
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/channel-icons/{channel_id}/{prefix}{icon_hash:0>32x}.{ext}'
    
    
def build_channel_group_icon_url_as(channel_id, icon_type, icon_hash, ext, size):
    """
    Returns the group channel's icon's image's url. If the channel has no icon, then returns `None`.
    
    Parameters
    ----------
    channel_id : `int`
        The channel's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    ext : `None | str`
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    
    size : `None | int`
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size, prefix == 'a_' and ext == 'webp')
    
    return f'{CDN_ENDPOINT}/channel-icons/{channel_id}/{prefix}{icon_hash:0>32x}.{ext}{end}'


def build_default_avatar_url(value):
    """
    Returns a default avatar's url.
    
    Parameters
    ----------
    value : `int`
        The value of the default avatar.
    
    Returns
    -------
    url : `str`
    """
    return f'{CDN_ENDPOINT}/embed/avatars/{value}.png'


def build_emoji_url(emoji_id, animated):
    """
    Returns the emoji's image's url. If the emoji is unicode emoji, then returns `None` instead.
    
    Parameters
    ----------
    emoji_id : `int`
        The emoji's identifier.
    
    animated : `bool`
        Whether the emoji is animated.
    
    Returns
    -------
    url : `None | str`
    """
    if emoji_id < (1 << 22):
        return None
    
    if animated:
         ext = 'gif'
    else:
         ext = 'png'
    
    return f'{CDN_ENDPOINT}/emojis/{emoji_id}.{ext}'


def build_emoji_url_as(emoji_id, animated, ext, size):
    """
    Returns the emoji's image's url. If the emoji is unicode emoji, then returns `None` instead.
    
    Parameters
    ----------
    emoji_id : `int`
        The emoji's identifier.
    
    animated : `bool`
        Whether the emoji is animated.
    
    ext : `None | str`
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`, `'avif'`.
        If emoji is animated, it can be `'gif'` as well.
    
    size : `None | int`
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    if emoji_id < (1 << 22):
        return None
    
    if ext is None:
        if animated:
            ext = 'gif'
        else:
            ext = 'png'
        
        add_animated_query_parameter = False
    else:
        if ext == 'avif':
            ext = 'webp'
        
        if animated:
            if ext not in VALID_ICON_FORMATS_EXTENDED:
                raise ValueError(
                    f'Extension must be one of {VALID_ICON_FORMATS_EXTENDED}, got {ext!r}.'
                )
            
            add_animated_query_parameter = ext == 'webp'
        else:
            if ext not in VALID_ICON_FORMATS:
                raise ValueError(
                    f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.'
                )
            
            add_animated_query_parameter = False
    
    end = _build_end(size, add_animated_query_parameter)
    
    return f'{CDN_ENDPOINT}/emojis/{emoji_id}.{ext}{end}'


def build_guild_badge_icon_url(guild_id, icon_type, icon_hash):
    """
    Returns the guild badge's icon's url. If the guild badge has no icon, then returns `None`.
    
    Parameters
    ----------
    guild_id : `int`
        The guild's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    Returns
    -------
    url : `None | str`
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/badge-icons/{guild_id}/{prefix}{icon_hash:0>32x}.{ext}'


def build_guild_badge_icon_url_as(guild_id, icon_type, icon_hash, ext, size):
    """
    Returns the guild badge's icon's url. If the guild badge has no icon, then returns `None`.
    
    Parameters
    ----------
    guild_id : `int`
        The guild's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    ext : `None | str`
        The extension of the image's url.
    
    size : `None | int`
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None | str`
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size, prefix == 'a_' and ext == 'webp')
    
    return f'{CDN_ENDPOINT}/badge-icons/{guild_id}/{prefix}{icon_hash:0>32x}.{ext}{end}'


def build_guild_banner_url(guild_id, icon_type, icon_hash):
    """
    Builds a guild's banner url.
    
    Parameters
    ----------
    guild_id : `int`
        The guild's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    Returns
    -------
    url : `None | str`
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/banners/{guild_id}/{prefix}{icon_hash:0>32x}.{ext}'


def build_guild_banner_url_as(guild_id, icon_type, icon_hash, ext, size):
    """
    Builds a guild's banner url with the given extension and size.
    
    Parameters
    ----------
    guild_id : `int`
        The guild's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    ext : `None | str`
        The extension of the image's url.
    
    size : `None | int`
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        - If `ext` or `size` were passed as an unexpected value.
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size, prefix == 'a_' and ext == 'webp')
    
    return f'{CDN_ENDPOINT}/banners/{guild_id}/{prefix}{icon_hash:0>32x}.{ext}{end}'


def build_guild_discovery_splash_url(guild_id, icon_type, icon_hash):
    """
    Builds a guild's discovery splash url.
    
    Parameters
    ----------
    guild_id : `int`
        The guild's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    Returns
    -------
    url : `None | str`
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/discovery-splashes/{guild_id}/{prefix}{icon_hash:0>32x}.{ext}'


def build_guild_discovery_splash_url_as(guild_id, icon_type, icon_hash, ext, size):
    """
    Builds a guild's discovery splash url with the given extension and size.
    
    Parameters
    ----------
    guild_id : `int`
        The guild's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    ext : `None | str`
        The extension of the image's url.
    
    size : `None | int`
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        - If `ext` or `size` were passed as an unexpected value.
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size, prefix == 'a_' and ext == 'webp')
    
    return f'{CDN_ENDPOINT}/discovery-splashes/{guild_id}/{prefix}{icon_hash:0>32x}.{ext}{end}'


def build_guild_home_splash_url(guild_id, icon_type, icon_hash):
    """
    Builds a guild's home splash url.
    
    Parameters
    ----------
    guild_id : `int`
        The guild's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    Returns
    -------
    url : `None | str`
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/home-headers/{guild_id}/{prefix}{icon_hash:0>32x}.{ext}'


def build_guild_home_splash_url_as(guild_id, icon_type, icon_hash, ext, size):
    """
    Builds a guild's home splash url with the given extension and size.
    
    Parameters
    ----------
    guild_id : `int`
        The guild's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    ext : `None | str`
        The extension of the image's url.
    
    size : `None | int`
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        - If `ext` or `size` were passed as an unexpected value.
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size, prefix == 'a_' and ext == 'webp')
    
    return f'{CDN_ENDPOINT}/home-headers/{guild_id}/{prefix}{icon_hash:0>32x}.{ext}{end}'


def build_guild_icon_url(guild_id, icon_type, icon_hash):
    """
    Builds a guild's icon url.
    
    Parameters
    ----------
    guild_id : `int`
        The guild's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    Returns
    -------
    url : `None | str`
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/icons/{guild_id}/{prefix}{icon_hash:0>32x}.{ext}'


def build_guild_icon_url_as(guild_id, icon_type, icon_hash, ext, size):
    """
    Builds a guild's icon url with the given extension and size.
    
    Parameters
    ----------
    guild_id : `int`
        The guild's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    ext : `None | str`
        The extension of the image's url.
    
    size : `None | int`
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        - If `ext` or `size` were passed as an unexpected value.
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size, prefix == 'a_' and ext == 'webp')
    
    return f'{CDN_ENDPOINT}/icons/{guild_id}/{prefix}{icon_hash:0>32x}.{ext}{end}'


def build_guild_invite_splash_url(guild_id, icon_type, icon_hash):
    """
    Builds a guild's invite splash url.
    
    Parameters
    ----------
    guild_id : `int`
        The guild's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    Returns
    -------
    url : `None | str`
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/splashes/{guild_id}/{prefix}{icon_hash:0>32x}.{ext}'


def build_guild_widget_json_url(guild_id):
    """
    Returns an url to the guild's widget data.
    
    Returns
    -------
    url : `str`
    """
    return  f'{API_ENDPOINT}/guilds/{guild_id}/widget.json'


def build_guild_widget_url(guild_id):
    """
    Returns the guild's widget image's url in `.png` format.
    
    Returns
    -------
    url : `str`
    """
    return f'{API_ENDPOINT}/guilds/{guild_id}/widget.png?style=shield'


def build_guild_widget_url_as(guild_id, style = 'shield'):
    """
    Returns the guild's widget image's url in `.png` format.
    
    Parameters
    ----------
    style : `str`
        The widget image's style. Can be any of: `'shield'`, `'banner1'`, `'banner2'`, `'banner3'`, `'banner4'`.
    
    Returns
    -------
    url : `str`
    
    Raises
    ------
    ValueError
        If `style` was not passed as any of the expected values.
    """
    if WIDGET_STYLE_RP.fullmatch(style) is None:
        raise ValueError(f'Invalid style: {style!r}')
    
    return f'{API_ENDPOINT}/guilds/{guild_id}/widget.png?style={style}'


def build_guild_invite_splash_url_as(guild_id, icon_type, icon_hash, ext, size):
    """
    Builds a guild's invite splash url with the given extension and size.
    
    Parameters
    ----------
    guild_id : `int`
        The guild's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    ext : `None | str`
        The extension of the image's url.
    
    size : `None | int`
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        - If `ext` or `size` were passed as an unexpected value.
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size, prefix == 'a_' and ext == 'webp')
    
    return f'{CDN_ENDPOINT}/splashes/{guild_id}/{prefix}{icon_hash:0>32x}.{ext}{end}'


def build_guild_vanity_invite_url(vanity_code):
    """
    Returns the guild's vanity invite's url.
    
    Parameters
    ----------
    vanity_code : `None | str`
        The guild's vanity invite core.
    
    Returns
    -------
    url : `None | str`
    """
    if (vanity_code is not None):
        return f'{INVITE_ENDPOINT}/{vanity_code}'


def build_invite_url(invite_code):
    """
    Returns the invite's url.
    
    Parameters
    ----------
    invite_code : `None | str`
        The invite's code.
    
    Returns
    -------
    url : `str`
    """
    return f'{INVITE_ENDPOINT}/{invite_code}'


def build_message_jump_url(guild_id, channel_id, message_id):
    """
    Builds a jump url to jump to a message in chat history.
    
    Parameters
    ----------
    guild_id : `int`
        The message's guild's identifier.
    
    channel_id : `int`
        The message's channel's identifier.
    
    message_id : `int`
        The message's identifier.
    
    Returns
    -------
    url : `str`
    """
    if guild_id:
        guild_id = str(guild_id)
    else:
        guild_id = '@me'
        
    return f'{DISCORD_ENDPOINT}/channels/{guild_id}/{channel_id}/{message_id}'


def build_name_plate_url(asset_path):
    """
    Builds name plate url.
    
    Parameters
    ----------
    asset_path : `str`
        Part to the name plate's asset.
    
    Returns
    -------
    url : `str`
    """
    return f'{DISCORD_ENDPOINT}/assets/collectibles/{asset_path}asset.webm'


def build_name_plate_url_as(asset_path, animated):
    """
    Builds name plate url.
    
    Parameters
    ----------
    asset_path : `str`
        Part to the name plate's asset.
    
    animated : `bool`
        Whether to return an url to an animated asset.
    
    Returns
    -------
    url : `str`
    """
    if animated:
        end = 'asset.webm'
    else:
        end = 'static.png'

    return f'{DISCORD_ENDPOINT}/assets/collectibles/{asset_path}{end}'


def build_role_icon_url(role_id, icon_type, icon_hash):
    """
    Returns the role's icon's image's url. If the role has no icon, then returns `None`.
    
    Parameters
    ----------
    role_id : `int`
        The role's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    Returns
    -------
    url : `None | str`
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/role-icons/{role_id}/{prefix}{icon_hash:0>32x}.{ext}'


def build_role_icon_url_as(role_id, icon_type, icon_hash, ext, size):
    """
    Returns the role's icon's image's url. If the role has no icon, then returns `None`.
    
    Parameters
    ----------
    role_id : `int`
        The role's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    ext : `None | str`
        The extension of the image's url.
    
    size : `None | int`
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size, prefix == 'a_' and ext == 'webp')
    
    return f'{CDN_ENDPOINT}/role-icons/{role_id}/{prefix}{icon_hash:0>32x}.{ext}{end}'


def build_scheduled_event_image_url(scheduled_event_id, icon_type, icon_hash):
    """
    Returns the scheduled event's image's url. If the scheduled event has no image, then returns `None`.
    
    Parameters
    ----------
    scheduled_event_id : `int`
        The scheduled event's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    Returns
    -------
    url : `None | str`
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/guild-events/{scheduled_event_id}/{prefix}{icon_hash:0>32x}.{ext}'


def build_scheduled_event_image_url_as(scheduled_event_id, icon_type, icon_hash, ext, size):
    """
    Returns the scheduled event's image's url. If the scheduled event has no image, then returns `None`.
    
    Parameters
    ----------
    scheduled_event_id : `int`
        The scheduled event's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    ext : `None | str`
        The extension of the image's url.
    
    size : `None | int`
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size, prefix == 'a_' and ext == 'webp')
    
    return f'{CDN_ENDPOINT}/guild-events/{scheduled_event_id}/{prefix}{icon_hash:0>32x}.{ext}{end}'


def build_scheduled_event_url(guild_id, scheduled_event_id):
    """
    Returns the scheduled event's url.
    
    Parameters
    ----------
    guild_id : `int`
        The guild's identifier where the scheduled event is.
    
    scheduled_event_id : `int`
        The scheduled event's identifier.
    
    Returns
    -------
    url : `str`
    """
    return f'{DISCORD_ENDPOINT}/events/{guild_id}/{scheduled_event_id}'


def build_soundboard_sound_url(sound_id):
    """
    Returns the url to the sound board sound.
    
    Parameters
    ----------
    sound_id : `int`
        Soundboard sound identifier.
    
    Returns
    -------
    url : `str`
    """
    return f'{CDN_ENDPOINT}/soundboard-sounds/{sound_id}'


def build_sticker_pack_banner_url(sticker_pack_banner_id):
    """
    Returns the sticker pack's banner's url.
    
    Parameters
    ----------
    sticker_pack_banner_id : `int`
        The sticker pack's banner's identifier.
    
    Returns
    -------
    url : `None | str`
    """
    if sticker_pack_banner_id:
        return f'{CDN_ENDPOINT}/app-assets/710982414301790216/store/{sticker_pack_banner_id}.png'


def build_sticker_pack_banner_url_as(sticker_pack_banner_id, ext, size):
    """
    Returns the sticker pack's banner's url.
    
    Parameters
    ----------
    sticker_pack_banner_id : `int`
        The sticker pack's banner's identifier.
    
    ext : `None | str`
        The extension of the banner's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    
    size : `None | int`
        The preferred minimal size of the banner's url.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    if not sticker_pack_banner_id:
        return
    
    end = _build_end(size, False)
    
    if ext is None:
        ext = 'png'
    
    else:
        if ext not in VALID_ICON_FORMATS:
            raise ValueError(
                f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.'
            )
    
    return f'{CDN_ENDPOINT}/app-assets/710982414301790216/store/{sticker_pack_banner_id}.{ext}{end}'


def build_sticker_url(sticker_id, sticker_format):
    """
    Returns the sticker's url.
    
    Parameters
    ----------
    sticker_id : `int`
        The sticker's identifier.
    
    sticker_format : ``StickerFormat``
        The sticker's format.
    
    Returns
    -------
    url : `None | str`
    """
    if sticker_format is StickerFormat.none:
        return None
    
    if sticker_format is StickerFormat.gif:
        endpoint = MEDIA_ENDPOINT
    else:
        endpoint = CDN_ENDPOINT
    
    return f'{endpoint}/stickers/{sticker_id}.{sticker_format.extension}'


def build_sticker_url_as(sticker_id, sticker_format, size, preview):
    """
    Returns the sticker's url.
    
    Parameters
    ----------
    sticker_id : `int`
        The sticker's identifier.
    
    sticker_format : ``StickerFormat``
        The sticker's format.
    
    size : `None | int`
        The preferred minimal size of the icon's url.
    
    preview : `bool`
        Whether preview url should be generated.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        If `size` was not passed as any of the expected values.
    """
    if sticker_format is StickerFormat.none:
        return None
    
    # Resolve size
    if size is None:
        end = ''
    else:
        if sticker_format is StickerFormat.lottie:
            end = ''
        else:
            if size in VALID_ICON_SIZES:
                end = f'?size={size}'
            else:
                raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size!r}.')
    
    # Resolve preview
    if preview:
        if sticker_format is StickerFormat.apng:
            end = f'{end}{"&" if end else "?"}passthrough=false'
    
    if sticker_format is StickerFormat.gif:
        endpoint = MEDIA_ENDPOINT
    else:
        endpoint = CDN_ENDPOINT
    
    return f'{endpoint}/stickers/{sticker_id}.{sticker_format.extension}{end}'


def build_team_icon_url(team_id, icon_type, icon_hash):
    """
    Returns the team's icon's url. If the team has no icon, then returns `None`.
    
    Parameters
    ----------
    team_id : `int`
        The team's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    Returns
    -------
    url : `None | str`
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/team-icons/{team_id}/{prefix}{icon_hash:0>32x}.{ext}'


def build_team_icon_url_as(team_id, icon_type, icon_hash, ext, size):
    """
    Returns the team's icon's url. If the team has no icon, then returns `None`.
    
    Parameters
    ----------
    team_id : `int`
        The team's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    ext : `None | str`
        The extension of the image's url.
    
    size : `None | int`
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size, prefix == 'a_' and ext == 'webp')
    
    return f'{CDN_ENDPOINT}/team-icons/{team_id}/{prefix}{icon_hash:0>32x}.{ext}{end}'


def build_user_avatar_url(user_id, icon_type, icon_hash):
    """
    Returns the user's avatar's url. If the user has no avatar, then returns `None`.
    
    Parameters
    ----------
    user_id : `int`
        The user's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    Returns
    -------
    url : `None | str`
    """
    if not icon_type.can_create_url():
        return None
   
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/avatars/{user_id}/{prefix}{icon_hash:0>32x}.{ext}'


def build_user_avatar_url_as(user_id, icon_type, icon_hash, ext, size):
    """
    Returns the user's avatar's url. If the user has no avatar, then returns `None`.
    
    Parameters
    ----------
    user_id : `int`
        The user's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    ext : `None | str`
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
        If the user has animated avatar, it can be `'gif'` as well.
    
    size : `None | int`
        The preferred minimal size of the avatar's url.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size, prefix == 'a_' and ext == 'webp')
    
    return f'{CDN_ENDPOINT}/avatars/{user_id}/{prefix}{icon_hash:0>32x}.{ext}{end}'


def build_user_avatar_url_for(user_id, guild_id, icon_type, icon_hash):
    """
    Returns the user's guild specific avatar. If the user has no guild local avatar, returns `None`.
    
    Parameters
    ----------
    user_id : `int`
        The user's identifier.
    
    guild_id : `int`
        The guild's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    Returns
    -------
    url : `None | str`
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/guilds/{guild_id}/users/{user_id}/avatars/{prefix}{icon_hash:0>32x}.{ext}'


def build_user_avatar_url_for_as(user_id, guild_id, icon_type, icon_hash, ext, size):
    """
    Returns the user's guild specific avatar. If the user has no guild local avatar, then returns `None`.
    
    Parameters
    ----------
    user_id : `int`
        The user's identifier.
    
    guild_id : `int`
        The guild's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    ext : `None | str`
        The extension of the image's url.
    
    size : `None | int`
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    if not icon_type.can_create_url():
        return None

    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size, prefix == 'a_' and ext == 'webp')
    
    return (
        f'{CDN_ENDPOINT}/guilds/{guild_id}/users/{user_id}/avatars/{prefix}{icon_hash:0>32x}.{ext}{end}'
    )


def build_user_banner_url(user_id, icon_type, icon_hash):
    """
    Returns the user's banner's url. If the user has no banner, then returns `None`.
    
    Parameters
    ----------
    user_id : `int`
        The user's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    Returns
    -------
    url : `None | str`
    """
    if not icon_type.can_create_url():
        return None
   
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/banners/{user_id}/{prefix}{icon_hash:0>32x}.{ext}'


def build_user_banner_url_as(user_id, icon_type, icon_hash, ext, size):
    """
    Returns the user's banner's url. If the user has no banner, then returns `None`.
    
    Parameters
    ----------
    user_id : `int`
        The user's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    ext : `None | str`
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
        If the user has animated banner, it can be `'gif'` as well.
    
    size : `None | int`
        The preferred minimal size of the banner's url.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size, prefix == 'a_' and ext == 'webp')
    
    return f'{CDN_ENDPOINT}/banners/{user_id}/{prefix}{icon_hash:0>32x}.{ext}{end}'


def build_user_banner_url_for(user_id, guild_id, icon_type, icon_hash):
    """
    Returns the user's guild specific banner. If the user has no guild local banner, returns `None`.
    
    Parameters
    ----------
    user_id : `int`
        The user's identifier.
    
    guild_id : `int`
        The guild's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    Returns
    -------
    url : `None | str`
    """
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/guilds/{guild_id}/users/{user_id}/banners/{prefix}{icon_hash:0>32x}.{ext}'


def build_user_banner_url_for_as(user_id, guild_id, icon_type, icon_hash, ext, size):
    """
    Returns the user's guild specific banner. If the user has no guild local banner, then returns `None`.
    
    Parameters
    ----------
    user_id : `int`
        The user's identifier.
    
    guild_id : `int`
        The guild's identifier.
    
    icon_type : ``IconType``
        The icon's type.
    
    icon_hash : `int`
        The icon's hash (uint128).
    
    ext : `None | str`
        The extension of the image's url.
    
    size : `None | int`
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None | str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    if not icon_type.can_create_url():
        return None

    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size, prefix == 'a_' and ext == 'webp')
    
    return (
        f'{CDN_ENDPOINT}/guilds/{guild_id}/users/{user_id}/banners/{prefix}{icon_hash:0>32x}.{ext}{end}'
    )


def build_webhook_url(webhook_id, webhook_token):
    """
    Returns the webhook's url.
    
    Parameters
    ----------
    webhook_id : `int`
        The webhook's identifier.
    
    webhook_token : `str`
        The webhook's token.
    
    Returns
    -------
    url : `str`
    """
    return f'{API_ENDPOINT}/webhooks/{webhook_id}/{webhook_token}'


def is_cdn_url(url):
    """
    Returns whether the given url a Discord content delivery network url.
    
    Parameters
    ----------
    url : `str`
        The url to check.
    
    Returns
    -------
    is_cdn_url : `bool`
    
    Examples
    --------
    Icons: `https://cdn.discordapp.com/...`
    Assets: `https://discord.com/...`
    Proxy service: `https://images-ext-1.discordapp.net/...`
    Attachments: `https://media.discordapp.net/...`
    ```
    """
    return (CDN_RP.match(url) is not None)


def is_media_url(url):
    """
    Returns whether the given url uses the discord's media content delivery network.
    
    Parameters
    ----------
    url : `str`
        The url to check.
    
    Returns
    -------
    is_media_url : `bool`
    """
    return url.startswith('https://media.discordapp.net/')


def parse_message_jump_url(message_url):
    """
    Parses the jump url of a message. On failure returns `0`-s.
    
    Parameters
    ----------
    message_url : `str`
        The message url to parse.
    
    Returns
    -------
    guild_id_and_channel_id_and_message_id : `(int, int, int)`
        The message's guild's, channel's and their own identifier.
    """
    parsed = MESSAGE_JUMP_URL_RP.fullmatch(message_url)
    if parsed is None:
        guild_id = 0
        channel_id = 0
        message_id = 0
    else:
        guild_id, channel_id, message_id = parsed.groups()
        if guild_id is None:
            guild_id = 0
        else:
            guild_id = int(guild_id)
        channel_id = int(channel_id)
        message_id = int(message_id)
    
    return guild_id, channel_id, message_id



__all__ = ()

from .preinstanced import ButtonStyle, SeparatorSpacingSize, TextInputStyle


BUTTON_STYLE_DEFAULT = ButtonStyle.blue
SEPARATOR_SPACING_SIZE_DEFAULT = SeparatorSpacingSize.small
TEXT_INPUT_STYLE_DEFAULT = TextInputStyle.short

DESCRIPTION_LENGTH_MIN = 0
DESCRIPTION_LENGTH_MAX = 1024

LABEL_LENGTH_MAX__BUTTON = 80
LABEL_LENGTH_MAX__LABEL = 45

PLACEHOLDER_LENGTH_MAX = 150

OPTIONS_LENGTH_MIN = 1
OPTIONS_LENGTH_MAX = 25

MAX_LENGTH_MIN = 0
MAX_LENGTH_MAX = 4000
MAX_LENGTH_DEFAULT = 0

MIN_LENGTH_MIN = 0
MIN_LENGTH_MAX = 4000
MIN_LENGTH_DEFAULT = 0

MAX_VALUES_MIN = 1
MAX_VALUES_MAX__ATTACHMENT_INPUT = 10
MAX_VALUES_MAX__SELECT = 25
MAX_VALUES_DEFAULT = 1

MIN_VALUES_MIN = 0
MIN_VALUES_MAX__ATTACHMENT_INPUT = 15
MIN_VALUES_MAX__SELECT = 25
MIN_VALUES_DEFAULT = 1

VALUE_LENGTH_MAX = 4000

URL_LENGTH_MAX = 512

CONTENT_LENGTH_MIN = 1
CONTENT_LENGTH_MAX = 4000
